from pdb import set_trace as T
import numpy as np
import random as rand

from neural_mmo.forge.blade.lib import material
from neural_mmo.forge.blade.io.stimulus.static import Stimulus
from neural_mmo.forge.blade.io.action import static as Action
from queue import PriorityQueue, Queue

from neural_mmo.forge.trinity.scripted import io, utils

def adjacentPos(pos):
   r, c = pos
   return [(r - 1, c), (r, c - 1), (r + 1, c), (r, c + 1)]

def inSight(dr, dc, vision):
    return (
          dr >= -vision and
          dc >= -vision and
          dr <= vision and
          dc <= vision)

def vacant(tile):
   Tile     = Stimulus.Tile
   occupied = io.Observation.attribute(tile, Tile.NEnts)
   matl     = io.Observation.attribute(tile, Tile.Index)

   return matl in material.Habitable and not occupied

def random(config, ob, actions):
   direction = rand.choice(Action.Direction.edges)
   actions[Action.Move] = {Action.Direction: direction}

def towards(direction):
   if direction == (-1, 0):
      return Action.North
   elif direction == (1, 0):
      return Action.South
   elif direction == (0, -1):
      return Action.West
   elif direction == (0, 1):
      return Action.East
   else:
      return rand.choice(Action.Direction.edges)

def pathfind(config, ob, actions, rr, cc):
   direction = aStar(config, ob, actions, rr, cc)
   direction = towards(direction)
   actions[Action.Move] = {Action.Direction: direction}

def meander(config, ob, actions):
   agent  = ob.agent
   Entity = Stimulus.Entity
   Tile   = Stimulus.Tile

   r = io.Observation.attribute(agent, Entity.R)
   c = io.Observation.attribute(agent, Entity.C)

   cands = []
   if vacant(ob.tile(-1, 0)):
      cands.append((-1, 0))
   if vacant(ob.tile(1, 0)):
      cands.append((1, 0))
   if vacant(ob.tile(0, -1)):
      cands.append((0, -1))
   if vacant(ob.tile(0, 1)):
      cands.append((0, 1))
   if not cands:
      return (-1, 0)

   direction = rand.choices(cands)[0]
   direction = towards(direction)
   actions[Action.Move] = {Action.Direction: direction}

def explore(config, ob, actions, spawnR, spawnC):
   vision = config.NSTIM
   sz     = config.TERRAIN_SIZE
   Entity = Stimulus.Entity
   Tile   = Stimulus.Tile

   agent  = ob.agent
   r      = io.Observation.attribute(agent, Entity.R)
   c      = io.Observation.attribute(agent, Entity.C)

   centR, centC = sz//2, sz//2
   vR, vC = centR-spawnR, centC-spawnC

   mmag = max(abs(vR), abs(vC))
   rr   = int(np.round(vision*vR/mmag))
   cc   = int(np.round(vision*vC/mmag))
   pathfind(config, ob, actions, rr, cc)

def evade(config, ob, actions, attacker):
   Entity = Stimulus.Entity

   sr     = io.Observation.attribute(ob.agent, Entity.R)
   sc     = io.Observation.attribute(ob.agent, Entity.C)

   gr     = io.Observation.attribute(attacker, Entity.R)
   gc     = io.Observation.attribute(attacker, Entity.C)

   rr, cc = (2*sr - gr, 2*sc - gc)

   pathfind(config, ob, actions, rr, cc)

def forageDijkstra(config, ob, actions, food_max, water_max, cutoff=100):
   vision = config.NSTIM
   Entity = Stimulus.Entity
   Tile   = Stimulus.Tile

   agent  = ob.agent
   food   = io.Observation.attribute(agent, Entity.Food)
   water  = io.Observation.attribute(agent, Entity.Water)

   best      = -1000 
   start     = (0, 0)
   goal      = (0, 0)

   reward    = {start: (food, water)}
   backtrace = {start: None}

   queue = Queue()
   queue.put(start)

   while not queue.empty():
      cutoff -= 1
      if cutoff <= 0:
         break

      cur = queue.get()
      for nxt in adjacentPos(cur):
         if nxt in backtrace:
            continue

         if not inSight(*nxt, vision):
            continue

         tile     = ob.tile(*nxt)
         matl     = io.Observation.attribute(tile, Tile.Index)
         occupied = io.Observation.attribute(tile, Tile.NEnts)

         if not vacant(tile):
            continue

         food, water = reward[cur]
         food  = max(0, food - 1)
         water = max(0, water - 1)

         if matl == material.Forest.index:
            food = min(food+food_max//2, food_max)
         for pos in adjacentPos(nxt):
            if not inSight(*pos, vision):
               continue

            tile = ob.tile(*pos)
            matl = io.Observation.attribute(tile, Tile.Index)
 
            if matl == material.Water.index:
               water = min(water+water_max//2, water_max)
               break

         reward[nxt] = (food, water)

         total = min(food, water)
         if total > best or (
                 total == best and max(food, water) > max(reward[goal])):
            best = total
            goal = nxt

         queue.put(nxt)
         backtrace[nxt] = cur

   while goal in backtrace and backtrace[goal] != start:
      goal = backtrace[goal]

   direction = towards(goal)
   actions[Action.Move] = {Action.Direction: direction}

def aStar(config, ob, actions, rr, cc, cutoff=100):
   Entity = Stimulus.Entity
   Tile   = Stimulus.Tile
   vision = config.NSTIM

   start = (0, 0)
   goal  = (rr, cc)

   if start == goal:
      return (0, 0)

   pq = PriorityQueue()
   pq.put((0, start))

   backtrace = {}
   cost = {start: 0}

   closestPos = start
   closestHeuristic = utils.l1(start, goal)
   closestCost = closestHeuristic

   while not pq.empty():
      # Use approximate solution if budget exhausted
      cutoff -= 1
      if cutoff <= 0:
         if goal not in backtrace:
            goal = closestPos
         break

      priority, cur = pq.get()

      if cur == goal:
         break

      for nxt in adjacentPos(cur):
         if not inSight(*nxt, vision):
            continue

         tile     = ob.tile(*nxt)
         matl     = io.Observation.attribute(tile, Tile.Index)
         occupied = io.Observation.attribute(tile, Tile.NEnts)

         #if not vacant(tile):
         #   continue

         if occupied:
            continue

         #Omitted water from the original implementation. Seems key
         if matl in material.Impassible:
            continue

         newCost = cost[cur] + 1
         if nxt not in cost or newCost < cost[nxt]:
            cost[nxt] = newCost
            heuristic = utils.lInfty(goal, nxt)
            priority = newCost + heuristic

            # Compute approximate solution
            if heuristic < closestHeuristic or (
                    heuristic == closestHeuristic and priority < closestCost):
               closestPos = nxt
               closestHeuristic = heuristic
               closestCost = priority

            pq.put((priority, nxt))
            backtrace[nxt] = cur

   #Not needed with scuffed material list above
   #if goal not in backtrace:
   #   goal = closestPos

   while goal in backtrace and backtrace[goal] != start:
      goal = backtrace[goal]

   return goal




from pdb import set_trace as T

class Tier:
   EASY   = 4
   NORMAL = 10
   HARD   = 25

class Diary:
   def __init__(self, config):
      self.achievements = []

      if config.game_system_enabled('Achievement'):
         self.achievements = [PlayerKills, Equipment, Exploration, Foraging]

      self.achievements = [a(config) for a in self.achievements]

   @property
   def stats(self):
      return [a.stats for a in self.achievements]

   def score(self, aggregate=True):
      score = [a.score for a in self.achievements]
      if score and aggregate:
         return sum(score)
      return score

   def update(self, realm, entity, aggregate=True, dry=False):
      scores = [a.update(realm, entity, dry) for a in self.achievements]
      if scores and aggregate:
         return sum(scores)
      return scores


class Achievement:
   def __init__(self, easy=None, normal=None, hard=None):
      self.progress = 0

      self.easy     = easy
      self.normal   = normal
      self.hard     = hard

   @property
   def stats(self):
      return self.__class__.__name__, self.progress

   @property
   def score(self):
      if self.hard and self.progress >= self.hard:
         return Tier.HARD
      elif self.normal and self.progress >= self.normal:
         return Tier.NORMAL
      elif self.easy and self.progress >= self.easy:
         return Tier.EASY
      return 0

   def update(self, value, dry):
      if value <= self.progress:
         return 0

      old = self.score

      if not dry:
         self.progress = value

      new = self.score

      if old == new:
         return 0

      return new - old
      
class PlayerKills(Achievement):
   def __init__(self, config):
      super().__init__(easy   = config.PLAYER_KILLS_EASY,
                       normal = config.PLAYER_KILLS_NORMAL,
                       hard   = config.PLAYER_KILLS_HARD)

   def update(self, realm, entity, dry):
      return super().update(entity.history.playerKills, dry)

class Equipment(Achievement):
   def __init__(self, config):
      super().__init__(easy   = config.EQUIPMENT_EASY,
                       normal = config.EQUIPMENT_NORMAL,
                       hard   = config.EQUIPMENT_HARD)

   def update(self, realm, entity, dry):
      return super().update(entity.loadout.defense, dry)

class Exploration(Achievement):
   def __init__(self, config):
      super().__init__(easy   = config.EXPLORATION_EASY,
                       normal = config.EXPLORATION_NORMAL,
                       hard   = config.EXPLORATION_HARD)

   def update(self, realm, entity, dry):
      return super().update(entity.history.exploration, dry)

class Foraging(Achievement):
   def __init__(self, config):
      super().__init__(easy   = config.FORAGING_EASY,
                       normal = config.FORAGING_NORMAL,
                       hard   = config.FORAGING_HARD)

   def update(self, realm, entity, dry):
      lvl = (entity.skills.fishing.level + entity.skills.hunting.level)/2.0
      return super().update(lvl, dry)




from pdb import set_trace as TT
import numpy as np

import torch
from torch import nn
from torch.nn.utils import rnn

from neural_mmo.forge.blade.io.stimulus.static import Stimulus

from neural_mmo.forge.ethyr.torch import policy
from neural_mmo.forge.ethyr.torch import io

class Base(nn.Module):
   def __init__(self, config):
      '''Base class for baseline policies

      Args:
         config: A Configuration object
      '''
      super().__init__()
      self.embed  = config.EMBED
      self.config = config

      self.output = io.Output(config)
      self.input  = io.Input(config,
            embeddings=policy.MixedDTypeInput,
            attributes=policy.SelfAttention)

      self.valueF = nn.Linear(config.HIDDEN, 1)

   def hidden(self, obs, state=None, lens=None):
      '''Abstract method for hidden state processing, recurrent or otherwise,
      applied between the input and output modules

      Args:
         obs: An observation dictionary, provided by forward()
         state: The previous hidden state, only provided for recurrent nets
         lens: Trajectory segment lengths used to unflatten batched obs
      ''' 
      raise NotImplementedError('Implement this method in a subclass')

   def forward(self, obs, state=None, lens=None):
      '''Applies builtin IO and value function with user-defined hidden
      state subnetwork processing. Arguments are supplied by RLlib
      ''' 
      entityLookup  = self.input(obs)
      hidden, state = self.hidden(entityLookup, state, lens)
      self.value    = self.valueF(hidden).squeeze(1)
      actions       = self.output(hidden, entityLookup)
      return actions, state

class Simple(Base):
   def __init__(self, config):
      '''Simple baseline model with flat subnetworks'''
      super().__init__(config)
      h = config.HIDDEN

      self.ent    = nn.Linear(2*h, h)
      self.conv   = nn.Conv2d(h, h, 3)
      self.pool   = nn.MaxPool2d(2)
      self.fc     = nn.Linear(h*6*6, h)

      self.proj   = nn.Linear(2*h, h)
      self.attend = policy.SelfAttention(self.embed, h)

   def hidden(self, obs, state=None, lens=None):
      #Attentional agent embedding
      agentEmb  = obs['Entity']
      selfEmb   = agentEmb[:, 0:1].expand_as(agentEmb)
      agents    = torch.cat((selfEmb, agentEmb), dim=-1)
      agents    = self.ent(agents)
      agents, _ = self.attend(agents)
      #agents = self.ent(selfEmb)

      #Convolutional tile embedding
      tiles     = obs['Tile']
      self.attn = torch.norm(tiles, p=2, dim=-1)

      w      = self.config.WINDOW
      batch  = tiles.size(0)
      hidden = tiles.size(2)
      #Dims correct?
      tiles  = tiles.reshape(batch, w, w, hidden).permute(0, 3, 1, 2)
      tiles  = self.conv(tiles)
      tiles  = self.pool(tiles)
      tiles  = tiles.reshape(batch, -1)
      tiles  = self.fc(tiles)

      hidden = torch.cat((agents, tiles), dim=-1)
      hidden = self.proj(hidden)
      return hidden, state

class Recurrent(Simple):
   def __init__(self, config):
      '''Recurrent baseline model'''
      super().__init__(config)
      self.lstm   = policy.BatchFirstLSTM(
            input_size=config.HIDDEN,
            hidden_size=config.HIDDEN)

   #Note: seemingly redundant transposes are required to convert between 
   #Pytorch (seq_len, batch, hidden) <-> RLlib (batch, seq_len, hidden)
   def hidden(self, obs, state, lens):
      #Attentional input preprocessor and batching
      lens = lens.cpu() if type(lens) == torch.Tensor else lens
      hidden, _ = super().hidden(obs)
      config    = self.config
      h, c      = state

      TB = hidden.size(0) #Padded batch of size (seq x batch)
      B  = len(lens)      #Sequence fragment time length
      T  = TB // B        #Trajectory batch size
      H  = config.HIDDEN  #Hidden state size

      #Pack (batch x seq, hidden) -> (batch, seq, hidden)
      hidden        = rnn.pack_padded_sequence(
                         input=hidden.view(B, T, H),
                         lengths=lens,
                         enforce_sorted=False,
                         batch_first=True)

      #Main recurrent network
      hidden, state = self.lstm(hidden, state)

      #Unpack (batch, seq, hidden) -> (batch x seq, hidden)
      hidden, _     = rnn.pad_packed_sequence(
                         sequence=hidden,
                         batch_first=True)

      return hidden.reshape(TB, H), state

class Attentional(Base):
   def __init__(self, config):
      '''Transformer-based baseline model'''
      super().__init__(config)
      self.agents = nn.TransformerEncoderLayer(d_model=config.HIDDEN, nhead=4)
      self.tiles  = nn.TransformerEncoderLayer(d_model=config.HIDDEN, nhead=4)
      self.proj   = nn.Linear(2*config.HIDDEN, config.HIDDEN)

   def hidden(self, obs, state=None, lens=None):
      #Attentional agent embedding
      agents    = self.agents(obs[Stimulus.Entity])
      agents, _ = torch.max(agents, dim=-2)

      #Attentional tile embedding
      tiles     = self.tiles(obs[Stimulus.Tile])
      self.attn = torch.norm(tiles, p=2, dim=-1)
      tiles, _  = torch.max(tiles, dim=-2)

      
      hidden = torch.cat((tiles, agents), dim=-1)
      hidden = self.proj(hidden)
      return hidden, state



class Observation:
   '''Unwraps observation tensors for use with scripted agents'''
   def __init__(self, config, obs):
      '''
      Args:
          config: A forge.blade.core.Config object or subclass object
          obs: An observation object from the environment
      '''
      self.config = config
      self.obs    = obs
      self.delta  = config.NSTIM
      self.tiles  = self.obs['Tile']['Continuous']
      self.agents = self.obs['Entity']['Continuous']
      self.n      = int(self.obs['Entity']['N'])

   def tile(self, rDelta, cDelta):
      '''Return the array object corresponding to a nearby tile
      
      Args:
         rDelta: row offset from current agent
         cDelta: col offset from current agent
      '''
      return self.tiles[self.config.WINDOW * (self.delta + rDelta) + self.delta + cDelta]

   @property
   def agent(self):
      '''Return the array object corresponding to the current agent'''
      return self.agents[0]

   @staticmethod
   def attribute(ary, attr):
      '''Return an attribute of a game object

      Args:
         ary: The array corresponding to a game object
         attr: A forge.blade.io.stimulus.static stimulus class
      '''
      return float(ary[attr.index])



'''Infrastructure layer for representing agent observations

Maintains a synchronized + serialized representation of agent observations in
flat tensors. This allows for fast observation processing as a set of tensor
slices instead of a lengthy traversal over hundreds of game properties.

Synchronization bugs are notoriously difficult to track down: make sure
to follow the correct instantiation protocol, e.g. as used for defining
agent/tile observations, when adding new types observations to the code'''

from pdb import set_trace as T
import numpy as np

from collections import defaultdict

from neural_mmo.forge.blade.io.stimulus import Static

class DataType:
   CONTINUOUS = np.float32
   DISCRETE   = np.int32

class Index:
   '''Lookup index of attribute names'''
   def __init__(self, prealloc):
      self.free  = {idx for idx in range(1, prealloc)}
      self.index = {}
      self.back  = {}

   def full(self):
      return len(self.free) == 0

   def remove(self, key):
      row = self.index[key]
      del self.index[key]
      del self.back[row]

      self.free.add(row)
      return row

   def update(self, key):
      if key in self.index:
         row = self.index[key]
      else:
         row = self.free.pop()
         self.index[key] = row
         self.back[row]  = key

      return row

   def get(self, key):
      return self.index[key]

   def teg(self, row):
      return self.back[row]

   def expand(self, cur, nxt):
      self.free.update({idx for idx in range(cur, nxt)})

class ContinuousTable:
   '''Flat tensor representation for a set of continuous attributes'''
   def __init__(self, config, obj, prealloc, dtype=DataType.CONTINUOUS):
      self.config = config
      self.dtype  = dtype
      self.cols   = {}
      self.nCols  = 0

      for (attribute,), attr in obj:
         self.initAttr(attribute, attr)

      self.data = self.initData(prealloc, self.nCols)

   def initAttr(self, key, attr):
      if attr.CONTINUOUS:
         self.cols[key] = self.nCols
         self.nCols += 1

   def initData(self, nRows, nCols):
      return np.zeros((nRows, nCols), dtype=self.dtype)

   def update(self, row, attr, val):
      col = self.cols[attr] 
      self.data[row, col] = val

   def expand(self, cur, nxt):
      data       = self.initData(nxt, self.nCols)
      data[:cur] = self.data

      self.data  = data
      self.nRows = nxt

   def get(self, rows, pad=None):
      data = self.data[rows]
      data[rows==0] = 0

      if pad is not None:
         data = np.pad(data, ((0, pad-len(data)), (0, 0)))

      return data

class DiscreteTable(ContinuousTable):
   '''Flat tensor representation for a set of discrete attributes'''
   def __init__(self, config, obj, prealloc, dtype=DataType.DISCRETE):
      self.discrete, self.cumsum = {}, 0
      super().__init__(config, obj, prealloc, dtype)

   def initAttr(self, key, attr):
      if not attr.DISCRETE:
         return

      self.cols[key]     =  self.nCols

      #Flat index
      attr               =  attr(None, None, 0, config=self.config)
      self.discrete[key] =  self.cumsum

      self.cumsum        += attr.max - attr.min + 1
      self.nCols         += 1

   def update(self, row, attr, val):
      col = self.cols[attr] 
      self.data[row, col] = val + self.discrete[attr]

class Grid:
   '''Flat representation of tile/agent positions'''
   def __init__(self, R, C):
      self.data = np.zeros((R, C), dtype=np.int)

   def zero(self, pos):
      r, c            = pos
      self.data[r, c] = 0
    
   def set(self, pos, val):
      r, c            = pos
      self.data[r, c] = val
 
   def move(self, pos, nxt, row):
      self.zero(pos)
      self.set(nxt, row)

   def window(self, rStart, rEnd, cStart, cEnd):
      crop = self.data[rStart:rEnd, cStart:cEnd].ravel()
      return list(filter(lambda x: x != 0, crop))
      
class GridTables:
   '''Combines a Grid + Index + Continuous and Discrete tables

   Together, these data structures provide a robust and efficient
   flat tensor representation of an entire class of observations,
   such as agents or tiles'''
   def __init__(self, config, obj, pad, prealloc=1000, expansion=2):
      self.grid       = Grid(config.TERRAIN_SIZE, config.TERRAIN_SIZE)
      self.continuous = ContinuousTable(config, obj, prealloc)
      self.discrete   = DiscreteTable(config, obj, prealloc)
      self.index      = Index(prealloc)

      self.nRows      = prealloc
      self.expansion  = expansion
      self.radius     = config.NSTIM
      self.pad        = pad

   def get(self, ent, radius=None, entity=False):
      if radius is None:
         radius = self.radius

      r, c = ent.pos
      cent = self.grid.data[r, c]
      assert cent != 0

      rows = self.grid.window(
            r-radius, r+radius+1,
            c-radius, c+radius+1)

      #Self entity first
      if entity:
         rows.remove(cent)
         rows.insert(0, cent)

      values = {'Continuous': self.continuous.get(rows, self.pad),
                'Discrete':   self.discrete.get(rows, self.pad)}

      if entity:
         ents = [self.index.teg(e) for e in rows]
         assert ents[0] == ent.entID
         return values, ents

      return values

   def update(self, obj, val):
      key, attr = obj.key, obj.attr
      if self.index.full():
         cur        = self.nRows
         self.nRows = cur * self.expansion

         self.index.expand(cur, self.nRows)
         self.continuous.expand(cur, self.nRows)
         self.discrete.expand(cur, self.nRows)

      row = self.index.update(key)
      if obj.DISCRETE:
         self.discrete.update(row, attr, val - obj.min)
      if obj.CONTINUOUS:
         self.continuous.update(row, attr, val)

   def move(self, key, pos, nxt):
      row = self.index.get(key)
      self.grid.move(pos, nxt, row)

   def init(self, key, pos):
      row = self.index.get(key)
      self.grid.set(pos, row)

   def remove(self, key, pos):
      self.index.remove(key)
      self.grid.zero(pos)

class Dataframe:
   '''Infrastructure wrapper class'''
   def __init__(self, config):
      self.config, self.data = config, defaultdict(dict)
      for (objKey,), obj in Static:
         self.data[objKey] = GridTables(config, obj, pad=obj.N(config))

   def update(self, node, val):
      self.data[node.obj].update(node, val)

   def remove(self, obj, key, pos):
      self.data[obj.__name__].remove(key, pos)

   def init(self, obj, key, pos):
      self.data[obj.__name__].init(key, pos)

   def move(self, obj, key, pos, nxt):
      self.data[obj.__name__].move(key, pos, nxt)

   def get(self, ent):
      stim = {}
     
      stim['Entity'], ents = self.data['Entity'].get(ent, entity=True)
      stim['Entity']['N']  = np.array([len(ents)], dtype=np.int32)

      ent.targets          = ents
      stim['Tile']         = self.data['Tile'].get(ent)

      return stim



from pdb import set_trace as T
import numpy as np

def unpack(vals, lens, dim=-1):
   '''Internal nD tensor pack utility'''
   if dim < 0:
      dim = len(vals.shape) + dim

   ret = []
   prefix = [slice(0, l) for l in vals.shape[1:dim]]
   suffix = [slice(0, l) for l in vals.shape[dim+1:]]
   for idx, l in enumerate(lens):
      middle = [slice(0, l)]
      inds = tuple([idx] + prefix + middle + suffix)
      e = vals[inds]
      ret.append(e)
   return ret

def pack(val):
   '''Internal nD tensor unpack utility'''
   shapes = np.array([e.shape for e in val])
   shape  = np.max(shapes, 0)

   #Must pad with zero (not -1) to satisfy embedding constraints
   dtype = val[0].dtype
   seq_tensor = np.zeros((len(val), *shape), dtype=dtype)
   seq_lens   = np.array([len(e) for e in val])

   for idx, tensor in enumerate(val):
      idx = tuple([idx] + [slice(0, i) for i in tensor.shape])
      seq_tensor[idx] = tensor
      
   return seq_tensor, seq_lens





from pdb import set_trace as T

import time
import os
import ray
import sys
import json
import argparse
import pickle
import numpy as np

from scipy.interpolate import interp1d
from collections import defaultdict

from functools import partial
from threading import Thread
from random import random

import bokeh
from bokeh.server.server import Server
from bokeh.models import ColumnDataSource
from bokeh.models import Band
from bokeh.models.widgets import RadioButtonGroup
from bokeh.layouts import layout

from bokeh.plotting import curdoc, figure, show
from bokeh.themes import Theme

from tornado import gen
from tornado.ioloop import IOLoop

from neural_mmo.forge.blade.lib.enums import Neon
from neural_mmo.forge.blade.systems.visualizer.config import *
from signal import signal, SIGTERM, SIGINT
from time import gmtime, strftime

def pickle_write(content, name, append=1):
    'function to open file, pickle dump, then close'
    f = open(name, "ab") if append else open(name, "wb")
    pickle.dump(content, f)
    f.close()

def pickle_read(name):
    'function to open file, pickle load, then close'
    f = open(name, "rb")
    ret = pickle.load(f)
    f.close()
    return ret

#Adapted from https://rigtorp.se/2011/01/01/rolling-statistics-numpy.html
def rolling_window(a, window, pad=True):
    a = np.array(a)
    if pad:
      a = np.pad(a, pad_width=(window-1, 0), mode='edge')
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

class Analytics:
    def __init__(self, config):
        """Visualizes a stream of data with threaded refreshing. To
        add items, initialize using 'keys' kwarg or add to packet in
        stream()
        Args:
            keys        : List of object names (str) to be displayed on market
            history_len : How far back to plot data
            scales      : Scales for visualization
            title       : Title of graph
            x           : Name of x axis data
            ylabel      : Name of y axis on plot
        """
        self.colors = []
        for color in [Neon.GREEN, Neon.CYAN, Neon.BLUE]:
           color = bokeh.colors.RGB(*(color.rgb))
           self.colors.append(color)
         
        # Market stores data in a dictionary of lists
        # MarketVisualizer stores the data for each view as a dictionary,
        # inside another dictionary with keys of each scale.

        self.history_len = config.HISTORY_LEN
        self.title       = config.TITLE

        self.ylabel      = config.YLABEL
        self.x           = config.XAXIS
        self.XAXIS       = config.XAXIS
        self.scales      = config.SCALES
        self.scale       = config.SCALES[0]
        self.title       = config.TITLE
        self.log         = config.LOG
        self.load        = config.LOAD_EXP
        self.filename    = config.NAME

        self.data        = defaultdict(list)
        self.dataSource  = {}
        self.keys        = 'lifetime'.split()
        for key in self.keys:
           self.dataSource[key] = [1, 2]
           self.dataSource[key+'_x'] = [1, 2]
           self.dataSource[key+'_lower'] = [1, 2]
           self.dataSource[key+'_upper'] = [1, 2]
           self.dataSource[key+'_smooth'] = [1, 2]

    def init(self, doc):
        # Source must only be modified through a stream
        self.source = ColumnDataSource(data=self.dataSource)

        # Enable theming
        theme = Theme(os.path.dirname(os.path.abspath(__file__)) + '/theme.yaml')
        doc.theme = theme
        self.doc = doc

        fig = figure(
           plot_width=960,
           plot_height=540,
           tools='xpan,xwheel_zoom, xbox_zoom, reset, save',
           title=self.title,
           x_axis_label=self.XAXIS,
           y_axis_label=self.ylabel)

        # Initialize plots
        for i, key in enumerate(self.keys):
            band = Band(
               source=self.source,
               base=key + '_x',
               lower=key+'_lower',
               upper=key+'_upper',
               level='annotation',
               line_color=self.colors[i],
               line_width=1.0,
               line_alpha=0.7,
               fill_color=self.colors[i],
               fill_alpha=0.3)
            fig.add_layout(band) 

            fig.line(
                source=self.source,
                x=key+'_x',
                y=key+'_smooth',
                color=self.colors[i],
                legend_label=key)

        def switch_scale(attr, old, new):
            """Callback for RadioButtonGroup to switch tick scale
            and refresh document
            Args:
                attr: variable to be changed, in this case 'active'
                old: old index of active button
                new: new index of active button
            """
            self.scale = self.scales[new]
            self.source.data = self.data[self.scale]

        self.timescales = RadioButtonGroup(labels=[str(scale) for scale in self.scales],
                                           active=self.scales.index(self.scale))
        self.timescales.on_change('active', switch_scale)

        self.structure = layout([[self.timescales], [fig]])
        self.doc.add_root(self.structure)

        self.fig = fig

    def update(self, packet):
      for key, val in packet.items():
         assert type(val) == list, 'packet value must be a list'
         self.data[key] += val

      #Sort by length for aesthetic rendering order
      self.keys = sorted(self.keys, key=lambda k: -len(self.data[k]))
        
    def resample(self, nOut=350, w=20):
      for key in list(self.data.keys()):
         #Padded sliding window
         val     = self.data[key]
         sliding = rolling_window(val, w)

         #Stats
         smooth  = np.mean(sliding, axis=1)
         std     = np.std(sliding, axis=1)

         #Add keys
         self.data[key+'_x']      = np.arange(len(val))
         self.data[key+'_smooth'] = smooth
         self.data[key+'_lower']  = smooth - std
         self.data[key+'_upper']  = smooth + std
 
      for key, val in self.data.items():
         #Interpolate
         nIn            = len(val)
         xIn            = np.arange(nIn)
         xOut           = np.linspace(0, nIn, nOut)
         f              = interp1d( 
                              x=xIn, y=val,
                              kind='nearest',
                              fill_value='extrapolate')

         self.data[key] = f(xOut)

    def stream(self):
        '''Wrapper function for source.stream to enable
        adding new items mid-stream. Overwrite graph
        with new figure if packet has different keys.
        Args:
            packet: dictionary of singleton lists'''

        # Stream items & append data if no new entries

        self.dataSource = dict(self.data.copy())
        if self.log:
           pickle_write(self.dataSource, self.filename)
        
        self.source.stream(self.dataSource, self.history_len)

        # Refreshes document to add new item
        self.doc.remove_root(self.structure)
        self.init(self.doc)


@ray.remote
class BokehServer:
    def __init__(self, middleman, config):
        """ Runs an asynchronous Bokeh data streaming server.
      
        Args:
           market : The market to visualize
           args   : Additional arguments
           kwargs : Additional keyword arguments
        """
        self.analytics = Analytics(config)
        self.middleman = middleman
        self.thread    = None

        server = Server(
                {'/': self.init},
                io_loop=IOLoop.current(),
                port=config.PORT,
                num_procs=1)

        server.start()
        self.server = server
        server.io_loop.add_callback(server.show, "/")
        server.io_loop.start()

    def init(self, doc):
        '''Initialize document and threaded update loop
        Args:
           doc: A Bokeh document
        '''
        self.analytics.init(doc)
        self.doc = doc

        self.thread = Thread(target=self.update, args=[])
        self.thread.start()
        self.started = True

    def update(self):
        '''Blocking update call to be run in a separate thread
        Ingests packets from a remote market and streams to Bokeh client'''
        self.n = 0
        while True:
            # Wait for thread to initialize
            time.sleep(0.05)
            if self.thread is None:
               continue 

            # Get remote market data
            if ray.get(self.middleman.getShutdown.remote()):
                self.middleman.setData.remote(self.analytics.data)
                sys.exit(0)

            packet = ray.get(self.middleman.getData.remote())
            if packet is None:
               continue

            self.analytics.update(packet)
            self.analytics.resample()

            # Stream to Bokeh client
            self.doc.add_next_tick_callback(partial(self.stream))

    @gen.coroutine
    def stream(self):
        '''Stream current data buffer to Bokeh client'''
        # visu is visualizer object
        self.analytics.stream()

@ray.remote
class Middleman:
    def __init__(self):
        '''Remote data buffer for two processes to dump and recv data.
        Interacts with Market and BokehServer.
        This is probably not safe'''
        self.data = None
        self.shutdown = 0

    def getData(self):
        '''Get data from buffer
        Returns:
           data: From buffer
        '''
        data = self.data
        self.data = None
        return data

    def setData(self, data):
        '''Set buffer data
        Args:
           data: To set buffer
        '''
        self.data = data.copy()

    # Notify remotes of visualizer shutdown
    def setShutdown(self):
        self.shutdown = 1

    # Retreive shutdown status
    def getShutdown(self):
        return self.shutdown

#class Market:
#    def __init__(self, items, middleman):
#        '''Dummy market emulator
#        Args: 
#           items     : List of item keys
#           middleman : A Middleman object'''
#        self.middleman = middleman
#        self.items     = items
#
#        self.keys = items
#        # Data contains market history for lowest values
#        self.data = {}
#
#        for i, key in enumerate(self.keys):
#            self.data[key] = 0
#
#    def update(self):
#        '''Updates market data and propagates to Bokeh server
#        # update data and send to middle man
#        Note: best to update all at once. Current version may cause bugs'''
#        for key, val in self.data.items():
#           self.data[key] = val + 0.2*(random() - 0.5)
#
#
#        # update if not shutting down visualizer
#        if not ray.get(middleman.getShutdown.remote()):
#            self.middleman.setData.remote(self.data)




from collections import deque

import numpy as np
from queue import PriorityQueue
from neural_mmo.forge.blade.systems import combat

class ExchangeEntry:
   def __init__(self, item, numBuy, numSell, maxBuyPrice, maxSellPrice):
      self.item = item
      self.numBuy = numBuy
      self.numSell = numSell
      self.maxBuyPrice = maxBuyPrice
      self.maxSellPrice = maxSellPrice

   #Hackey str get. Only for printing
   @property
   def itemName(self):
      return str(self.item)[17:-2]

   @property
   def worth(self):
      return (self.numBuy + self.numSell)*self.maxBuyPrice

   def __lt__(self, other):
      return True

   def __eq__(self, other):
      return False

class StatTraker:
   def __init__(self, maxLen=2048):
      self.lenTrak = deque(maxlen=maxLen)

   def update(self, pcs, npcs, exchange):
      self.pcs = pcs
      self.npcs = npcs
      
      self.lenTrak.append(len(pcs))
      #Update exchange
      self.updateExchange(exchange)
   
   def updateExchange(self, exchange):
      self.exchange = PriorityQueue()
      buyOffers = exchange.buyOffers
      sellOffers = exchange.sellOffers
      buyKeys, sellKeys = buyOffers.keys(), sellOffers.keys()
      exchangeKeys = list(set(list(buyKeys) + list(sellKeys)))

      for key in exchangeKeys:
         keyBuys, keySells  = buyOffers[key], sellOffers[key]
         topBuy  = keyBuys.peek()
         topSell = keySells.peek()

         if topBuy is not None:
            numBuy, maxBuyPrice   = topBuy.quantLeft, topBuy.itemPrice
            item = topBuy.item
         else:
            numBuy, maxBuyPrice = 0, 0

         if topSell is not None:
            numSell, minSellPrice = topSell.quantLeft, topSell.itemPrice
            item = topSell.item
         else:
            numSell, minSellPrice = 0, 0

         #Compute total buy value
         totalBuy = 0
         for e in keyBuys.queue:
            totalBuy += e.coffer

         entry = ExchangeEntry(item, numBuy, numSell, 
               maxBuyPrice, minSellPrice)

         self.exchange.put(entry, totalBuy)
 
   @property
   def numEntities(self):
      return np.asarray(list(self.lenTrak))

class StatBlock:
   def __init__(self, entity):
      self.timeAlive = entity.timeAlive
      self.level   = combat.combatLevel(entity.skills)
      self.defense = entity.skills.defense.level
      self.melee   = entity.skills.melee.level
      self.ranged  = entity.skills.ranged.level
      
   



from pdb import set_trace as T

class StimHook:                                                               
   def __init__(self, meta, config):                                          
      self.meta = meta                                                        
      self.config = config                                                    
                                                                              
      self.inputs(meta, config)                                               
                                                                              
   def inputs(self, cls, config):                                             
      for _, c in cls:                                                     
         self.__dict__[c.name] = c(config)                                    
                                                                              
   def outputs(self, config):                                                 
      data = {}                                                               
      for name, cls in self.meta:                                             
         attr           = self.__dict__[cls.name]
         data[cls.name] = attr.packet()                                           
                                                                              
      return data                                                             
                                                                              
   def packet(self):                                                          
      return self.outputs(self.config)



#Iterates through all items and hooks them to their respective skills
from neural_mmo.forge.blade.item import item
from neural_mmo.forge.blade.lib import utils

#Filled in at runtime with items
class ItemList:
   items = []

def hook():
   cls = item.Item
   for e in utils.terminalClasses(cls):
      skill = e.createSkill
      if skill is not None:
         if type(skill.skillItems) != list:
            skill.skillItems = []
         skill.skillItems += [e]
         ItemList.items += [e]



from pdb import set_trace as T
import numpy as np

import torch
from torch import nn
from torch.nn import functional as F

#NaN grad errors here usually mean too small
#of a batch for good advantage estimation
#(or dummy fixed action selection)
def advantage(returns, val):
   '''Computes a mean centered advantage function
   using a value function baseline.

   Args:
      returns: Tensor of trajectory returns
      vals: Tensor of value function outpus

   Returns:
      advantage: Advantage estimate tensor
   '''
   A = returns - val
   adv = A
   adv = (A - A.mean())
   #adv = (A - A.mean()) / (1e-4 + A.std())
   #adv = (A - A.mean()) / (1e-8 + A.std())
   adv = adv.detach()
   return adv

def policyLoss(logProb, atn, adv):
   '''Policy gradient loss
   
   Args:
      logProb: Log probability tensor
      atn: Action index tensor
      adv: Advantage estimate tensor

   Returns:
      loss: Mean policy gradient loss
   '''
   pgLoss = -logProb.gather(1, atn.view(-1, 1))
   return (pgLoss * adv).mean()

def valueLoss(val, returns):
   '''Value function loss

   Args:
      val: Value tensor
      returns: Return tensor

   Returns:
      value: Mean value loss
   '''
   return (0.5 * (val - returns) **2).mean()

def entropyLoss(prob, logProb):
   '''Entropy computation
   
   Args:
      prob    : Probability tensor
      logProb : Log probability tensor
   
   Returns:
      entropy: Mean entropy 
   '''
   loss = (prob * logProb)
   loss[torch.isnan(loss)] = 0
   return loss.sum(1).mean()

#Assumes pi is already -inf padded
def PG(pi, atn, val, returns):
   '''Computes losses for the policy gradient algorithm
   
   Args:
      pi     : Logit tensor (network outputs, pre softmax)
      atn    : Index tensor of selected actions
      val    : Value tensor
      return : Return tensor 
   
   Returns:
      polLoss : Policy loss
      entLoss : Entropy loss
   '''
   prob = [F.softmax(e, dim=-1) for e in pi]
   logProb = [F.log_softmax(e, dim=-1) for e in pi]

   prob = torch.nn.utils.rnn.pad_sequence(prob, batch_first=True)
   logProb = torch.nn.utils.rnn.pad_sequence(logProb, batch_first=True)

   adv = advantage(returns, val)
   polLoss = policyLoss(logProb, atn, adv)
   entLoss = entropyLoss(prob, logProb)
   valLoss = valueLoss(val, returns)
   return polLoss, valLoss, entLoss

#Not currently used because Vanilla PG is equivalent with 1 update/minbatch
class PPO(nn.Module):
   def forward(self, pi, v, atn, returns):
      prob, logProb = F.softmax(pi, dim=1), F.log_softmax(pi, dim=1)
      atnProb = prob.gather(1, atn.view(-1, 1))

      #Compute advantage
      A = returns - v
      adv = (A - A.mean()) / (1e-4 + A.std())
      adv = adv.detach()

      #Clipped ratio loss
      prob    = F.softmax(pi, dim=1)
      probOld = F.softmax(piOld, dim=1)
      logProb = F.log_softmax(pi, dim=1)

      atnProb    = prob.gather(1, atn)
      atnProbOld = probOld.gather(1, atn)

      ratio = atnProb / (atnProbOld + 1e-6)
      surr1 = ratio*adv
      surr2 = torch.clamp(ratio, min=1. - self.clip, max=1. + self.clip) * adv 
      policyLoss = -torch.min(surr1, surr2)

      #Compute value loss
      valueLoss = (0.5 * (v - returns) **2).mean()

      #Compute entropy loss
      entropyLoss = (prob * logProb).sum(1).mean()

      return policyLoss, valueLoss, entropyLoss



from pdb import set_trace as T

import inspect
from collections import defaultdict

from neural_mmo.forge.blade.lib.enums import Neon, Solid

class Iterable(type):
   def __iter__(cls):
      stack = list(cls.__dict__.items())
      while len(stack) > 0:
         name, attr = stack.pop()
         if name.startswith('__'):
            continue
         if not inspect.isclass(attr) and type(attr) == classmethod:
            continue
         yield name, attr
  
class I(metaclass=Iterable):
   pass

class Theme(metaclass=Iterable):
   @classmethod
   def dict(cls):
     data = defaultdict(dict)

     for group, c in Theme:
        for attr, val in c:
           data[group][attr] = val

     for group, c in cls:
        for attr, val in c:
           data[group][attr] = val

     return dict(data)

   class Figure(I):
      outline_line_cap = "square"
      outline_line_width    = 2.0

   class Grid(I):
      grid_line_dash  = [6, 4]
      grid_line_width = 1.1

   class Title(I):
      text_font_size = "20pt"
      align          = "center"

   class Axis(I):
      axis_label_text_font_size  = "16pt"
      major_label_text_font_size = "14pt"
      major_tick_line_cap   = None
      minor_tick_line_cap   = None
      major_tick_line_width = 2.0
      minor_tick_line_width = 1.0
      
   class Legend(I):
      background_fill_alpha = 0.75
      border_line_width     = 0
      border_line_alpha     = 1.0
      label_text_font_size  = "14pt"
      click_policy          = "hide"
      location              = "bottom_right"

   class Line(I):
      line_width = 2.5
      line_cap   = 'square' 

   class Circle(I):
      fill_alpha = 0.3
      line_width = 1.0
      size       = 5.0

   class Quad(I):
      fill_alpha = 0.5

   class VArea(I):
      fill_alpha = 0.3

   class Patch(I):
      fill_alpha = 0.3
   

class Web(Theme):
   def __init__(self, config):
      self.index  = config.PATH_THEME_WEB
      self.colors = Neon.color12()
   
   class Figure(I):
      background_fill_color = "#000000"
      border_fill_color     = "#000e0e"
      outline_line_color    = "#005050"

   class Grid(I):
      grid_line_color = "#005050"
      grid_line_alpha = 0.0

   class Title(I):
      text_color     = "#00bbbb"

   class Axis(I):
      axis_line_color            = "#005050"
      axis_label_text_color      = "#00bbbb"
      major_label_text_color     = "#005050"
      major_tick_line_color      = "#005050"
      minor_tick_line_color      = "#005050"

   class Legend(I):
      background_fill_color = "#000e0e"
      inactive_fill_color   = "#000e0e"
      border_line_color     = "#005050"
      label_text_color      = "#00bbbb"


class Publication(Theme):
   def __init__(self, config):
      self.index  = config.PATH_THEME_PUB
      self.colors = Solid.color10()
 
   class Figure(I):
      background_fill_color = "#ffffff"
      border_fill_color     = "#ffffff"
      outline_line_color    = "#000000"

   class Grid(I):
      grid_line_color = "#505050"
      grid_line_alpha = 0.0

   class Title(I):
      text_color = "#000000"

   class Axis(I):
      axis_line_color        = "#000000"
      axis_label_text_color  = "#000000"
      major_label_text_color = "#000000"
      major_tick_line_color  = "#000000"
      minor_tick_line_color  = "#000000"

   class Legend(I):
      background_fill_color = "#bbbbbb"
      inactive_fill_color   = "#bbbbbb"
      border_line_color     = "#000000"
      label_text_color      = "#000000"



from pdb import set_trace as T
import numpy as np

from signal import signal, SIGINT
import sys, os, json, pickle, time
import ray

from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from twisted.python import log
from twisted.web.server import Site
from twisted.web.static import File

from autobahn.twisted.websocket import WebSocketServerFactory, \
    WebSocketServerProtocol
from autobahn.twisted.resource import WebSocketResource

def sign(x):
    return int(np.sign(x))

def softmax(x):
    exp = np.exp(x)
    return exp / np.sum(exp)

def move(orig, targ):
    ro, co = orig
    rt, ct = targ
    dr = rt - ro
    dc = ct - co
    if abs(dr) > abs(dc):
        return ro + sign(dr), co
    elif abs(dc) > abs(dr):
        return ro, co + sign(dc)
    else:
        return ro + sign(dr), co + sign(dc)

class GodswordServerProtocol(WebSocketServerProtocol):
    def __init__(self):
        super().__init__()
        print("Created a server")
        self.frame  = 0

        #"connected" is already used by WSSP
        self.isConnected = False
        self.pos = [512, 512]
        self.cmd = None

    def onOpen(self):
        print("Opened connection to server")

    def onClose(self, wasClean, code=None, reason=None):
        self.isConnected = False
        print('Connection closed')

    def connectionMade(self):
        super().connectionMade()
        self.factory.clientConnectionMade(self)

    def connectionLost(self, reason):
        super().connectionLost(reason)
        self.factory.clientConnectionLost(self)

    #Not used without player interaction
    def onMessage(self, packet, isBinary):
        print("Server packet", packet)
        packet    = packet.decode()
        _, packet = packet.split(';') #Strip headeer
        r, c, cmd = packet.split(' ') #Split camera coords
        if len(cmd) == 0:
            cmd = None

        self.pos = [int(r), int(c)]
        self.cmd = cmd

        self.isConnected = True

    def onConnect(self, request):
        print("WebSocket connection request: {}".format(request))
        realm = self.factory.realm
        self.realm = realm
        self.frame += 1

        data = self.serverPacket()
        self.sendUpdate()

    def serverPacket(self):
        data = self.realm.packet
        return data

    def sendUpdate(self):
        data = self.serverPacket()

        packet               = {}
        packet['resource']   = data['resource']
        packet['player']     = data['player']
        packet['npc']        = data['npc']
        packet['pos']        = data['pos']
        packet['wilderness'] = data['wilderness']

        config = data['config']

        print('Is Connected? : {}'.format(self.isConnected))
        if not self.isConnected:
            packet['map']    = data['environment']
            packet['border'] = config.TERRAIN_BORDER
            packet['size']   = config.TERRAIN_SIZE

        if 'overlay' in data:
           packet['overlay'] = data['overlay']
           print('SENDING OVERLAY: ', len(packet['overlay']))

        packet = json.dumps(packet).encode('utf8')
        self.sendMessage(packet, False)

class WSServerFactory(WebSocketServerFactory):
    def __init__(self, ip, realm, step):
        super().__init__(ip)
        self.realm, self.step = realm, step
        self.time = time.time()
        self.clients = []

        self.pos = [40, 40]
        self.cmd = None
        self.tickRate = 0.6
        self.tick = 0

        self.step(self.pos, self.cmd)
        lc = LoopingCall(self.announce) #If this misses a tick, it waits for a whole cycle
        lc.start(self.tickRate)

    def announce(self):
        self.tick += 1
        uptime = np.round(self.tickRate*self.tick, 1)
        print('Wall Clock: ', time.time() - self.time, 'Uptime: ', uptime, ', Tick: ', self.tick)
        self.time = time.time()

        for client in self.clients:
            client.sendUpdate()
            if client.pos is not None:
                self.pos = client.pos
                self.cmd = client.cmd

        self.step(self.pos, self.cmd)

    def clientConnectionMade(self, client):
        self.clients.append(client)

    def clientConnectionLost(self, client):
        self.clients.remove(client)

class Application:
   def __init__(self, realm, step):
      signal(SIGINT, self.kill)
      self.realm = realm
      log.startLogging(sys.stdout)
      port = 8080

      factory          = WSServerFactory(u'ws://localhost:' + str(port), realm, step)
      factory.protocol = GodswordServerProtocol 
      resource         = WebSocketResource(factory)

      # We server static files under "/" and our WebSocket 
      # server under "/ws" (note that Twisted uses bytes 
      # for URIs) under one Twisted Web Site
      root = File(".")
      root.putChild(b"ws", resource)
      site = Site(root)

      reactor.listenTCP(port, site)
      reactor.run()

   def kill(*args):
      print("Killed by user")
      reactor.stop()
      os._exit(0)




from pdb import set_trace as T

import numpy as np
import sys
import os
 
from neural_mmo.forge.blade.systems.visualizer import plot, theme
from neural_mmo.forge.blade.lib.log import Quill
from neural_mmo.forge.trinity import formatting

import bokeh, jinja2

class BokehServer:
   '''Interactive dashboard for visualizing evaluation logs

   This dashboard is created automatically from the data and plot types
   collected in Env.log. It supports both web and publication themes.'''
   def __init__(self, config):
      self.config = config

      server = bokeh.server.server.Server(
         {'/': self.visualize},
         port=self.config.VIS_PORT,
         num_procs=1)

      server.start()
      server.io_loop.add_callback(server.show, '/')
      server.io_loop.start()

   def load_theme(self, doc):
      '''Load dashboard theme'''
      config = self.config
      if config.VIS_THEME == 'web':
         preset = theme.Web(config)
      elif config.VIS_THEME == 'publication':
         preset = theme.Publication(config)
      else: 
         sys.exit('THEME_NAME invalid: web or publication')

      self.colors = [bokeh.colors.RGB(*(c.rgb)) for c in preset.colors]

      doc.theme = bokeh.themes.Theme(json={'attrs': preset.dict()})
      with open(preset.index) as f:
         doc.template = jinja2.environment.Template(f.read())

   def load_data(self):
      train_data, eval_data = {}, {}
      config = self.config

      train_path = config.PATH_TRAINING_DATA.format(config.MODEL)
      if os.path.exists(train_path):
         train_data = np.load(train_path, allow_pickle=True).tolist()['logs']
      else:
         print('VISUALIZE: No training logs for specified config')

      eval_path = config.PATH_EVALUATION
      if os.path.exists(eval_path):
         eval_data = np.load(eval_path, allow_pickle=True).tolist()
      else:
         print('VISUALIZE: No evaluation logs for specified config')

      return train_data, eval_data

   def resample(self, data, n):
      config  = self.config
      n      -= 1

      for track, stats in data.items():
         for key, vals in stats.items():
            llen   = len(vals) - 1
            window =  llen // n - 1
            sz     = window * n
            ary    = np.array(vals[1:sz+1]).reshape(n, -1)

            first = vals[0]
            mid   = ary[:-1].mean(1).tolist()
            last  = np.mean(ary[-1].tolist() + vals[sz+1:])
         
            ary        = np.array([first] + mid + [last])
            mmin, mmax = np.min(ary), np.max(ary)
            ary        = (ary - mmin) / (mmax - mmin)

            data[track][key] = ary.tolist()
         data[track]['x'] = [1] + (1 + window*np.arange(1, n)).tolist() + [len(vals)]
 
   def visualize(self, doc):
      '''Build dashboard'''
      config = self.config
      self.load_theme(doc)

      #Load data
      train_data, eval_data = self.load_data()
      if len(train_data) > 0:
         self.resample(train_data, config.TRAIN_DATA_RESAMPLE)
         train_data = {('Training', (-1,)): train_data}
      if len(eval_data) > 0:
         for line in formatting.table_stats(eval_data['Stats']):
            print(line)
         eval_data  = eval_data['Log'][0]

      layout = []
      for (key, plots), blob in {**train_data, **eval_data}.items():
         row, n = [], len(plots)

         #Left title
         titleObj = bokeh.models.Title(text=key, align='center', vertical_align='middle')
         fig, _   = plot.blank(60, config.VIS_HEIGHT)
         fig.add_layout(titleObj, 'left')
         row.append(fig)

         #Monochrome or full palette
         colors = self.colors
         if config.VIS_THEME == 'publication' and len(blob)==1:
            colors = [bokeh.colors.RGB(0, 0, 0)]
 
         #Draw plot row
         for idx, p in enumerate(plots):
            fig, legend = self.plot(key, blob, p, colors, idx, n)
            row.append(fig)
         
         #Right legend
         fig, legend = plot.blank(
               config.VIS_LEGEND_WIDTH + config.VIS_LEGEND_OFFSET,
               config.VIS_HEIGHT, blob, key, colors)
         fig.outline_line_width = 0
         fig.min_border_right   =0
         items  = list(legend.items())
         legend = bokeh.models.Legend(items=items, location='center')
         legend.label_width = config.VIS_LEGEND_WIDTH
         fig.add_layout(legend, 'right')
         row.append(fig)

         layout.append(row)

      
      doc.add_root(bokeh.layouts.layout(layout))
      return doc


   def plot(self, ylabel, blob, plot, colors, idx, n):
      config = self.config
      Plot   = Quill.plot(plot)
      title  = Plot.__name__

      #Correct dimensions
      width  = ((config.VIS_WIDTH - config.VIS_LEGEND_WIDTH
            - config.VIS_LEGEND_OFFSET - config.VIS_TITLE_OFFSET)  // n)
      height = config.VIS_HEIGHT
      aspect = height / width

      #Interactive plot element
      TOOLTIPS = [
         ("Track", "$name"),
         ("(x,y)", "($x, $y)"),
      ]

      #Make figure
      fig = bokeh.plotting.figure(
         plot_width=width,
         plot_height=height,
         x_range=bokeh.models.ranges.DataRange1d(range_padding=0.1*aspect),
         y_range=bokeh.models.ranges.DataRange1d(range_padding=0.1),
         tools='hover, pan, box_zoom, wheel_zoom, reset, save',
         active_drag='pan',
         active_scroll='wheel_zoom',
         active_inspect='hover',
         tooltips=TOOLTIPS,
         title_location='above',
         y_axis_label=ylabel,
         x_axis_location='below',
         min_border_right=config.VIS_BORDER_WIDTH,
         min_border_left=config.VIS_BORDER_WIDTH,
         min_border_top=config.VIS_BORDER_HEIGHT,
         min_border_bottom=config.VIS_BORDER_HEIGHT)

      #Extra options
      fig.axis.axis_label_text_font_style = 'bold'

      #Toolbars
      fig.toolbar.logo = None
      if not config.VIS_TOOLS:
         fig.toolbar_location = None
        
      #Draw glyphs
      legend = Plot(config, fig, ylabel, len(blob)).render(blob, colors)

      #Adjust for theme
      if config.VIS_THEME == 'web':
         fig.title.text = '{} vs. {}'.format(
               fig.yaxis.axis_label, fig.xaxis.axis_label)
         fig.yaxis.axis_label = None
         fig.xaxis.axis_label = None

      return fig, legend




from pdb import set_trace as T
import numpy as np

from collections import defaultdict, Mapping
from typing import Dict, Callable

from neural_mmo.forge.blade import core
from neural_mmo.forge.blade.lib.enums import Palette
from neural_mmo.forge.blade.entity.npc import NPC
from neural_mmo.forge.blade.entity import Player
from neural_mmo.forge import trinity

from neural_mmo.forge.blade.io.stimulus import Static

def prioritized(entities: Dict, merged: Dict):
   '''Sort actions into merged according to priority'''
   for idx, actions in entities.items():
      for atn, args in actions.items():
         merged[atn.priority].append((idx, (atn, args.values())))
   return merged

class EntityGroup(Mapping):
   def __init__(self, config, realm):
      self.dataframe = realm.dataframe
      self.config    = config

      self.entities  = {}
      self.dead      = {}

   def __len__(self):
      return len(self.entities)

   def __contains__(self, e):
      return e in self.entities

   def __getitem__(self, key):
      return self.entities[key]
   
   def __iter__(self):
      yield from self.entities

   def items(self):
      return self.entities.items()

   @property
   def corporeal(self):
      return {**self.entities, **self.dead}

   @property
   def packet(self):
      return {k: v.packet() for k, v in self.corporeal.items()}

   def reset(self):
      for entID, ent in self.entities.items():
         self.dataframe.remove(Static.Entity, entID, ent.pos)

      self.spawned  = False
      self.entities = {}
      self.dead     = {}

   def add(iden, entity):
      assert iden not in self.entities
      self.entities[iden] = entity

   def remove(iden):
      assert iden in self.entities
      del self.entities[iden] 

   def spawn(self, entity):
      pos, entID = entity.pos, entity.entID
      self.realm.map.tiles[pos].addEnt(entity)
      self.entities[entID] = entity
 
   def cull(self):
      self.dead = {}
      for entID in list(self.entities):
         player = self.entities[entID]
         if not player.alive:
            r, c  = player.base.pos
            entID = player.entID
            self.dead[entID] = player

            self.realm.map.tiles[r, c].delEnt(entID)
            del self.entities[entID]
            self.realm.dataframe.remove(Static.Entity, entID, player.pos)

      return self.dead

   def update(self, actions):
      for entID, entity in self.entities.items():
         entity.update(self.realm, actions)

class NPCManager(EntityGroup):
   def __init__(self, config, realm):
      super().__init__(config, realm)
      self.realm = realm
      self.idx   = -1
 
   def spawn(self):
      if not self.config.game_system_enabled('NPC'):
         return

      for _ in range(self.config.NPC_SPAWN_ATTEMPTS):
         if len(self.entities) >= self.config.NMOB:
            break

         center = self.config.TERRAIN_CENTER
         border = self.config.TERRAIN_BORDER
         r, c   = np.random.randint(border, center+border, 2).tolist()
         if self.realm.map.tiles[r, c].occupied:
            continue

         npc = NPC.spawn(self.realm, (r, c), self.idx)
         if npc: 
            super().spawn(npc)
            self.idx -= 1

   def actions(self, realm):
      actions = {}
      for idx, entity in self.entities.items():
         actions[idx] = entity.decide(realm)
      return actions
       
class PlayerManager(EntityGroup):
   def __init__(self, config, realm, identify: Callable):
      super().__init__(config, realm)
      self.identify = identify
      self.realm    = realm

      self.palette = Palette(config.NPOP)
      self.idx     = 1

   def spawnIndividual(self, r, c):
      pop, name = self.identify()
      color     = self.palette.color(pop)
      player    = Player(self.realm, (r, c), self.idx, pop, name, color)
      super().spawn(player)
      self.idx += 1

   def spawn(self):
      if self.config.SPAWN == self.config.SPAWN_CONCURRENT:
         if self.spawned:
            return 

         self.spawned = True
         for r, c in self.config.SPAWN():
            assert not self.realm.map.tiles[r, c].occupied
            self.spawnIndividual(r, c)
         return
          
      #MMO-style spawning
      for _ in range(self.config.PLAYER_SPAWN_ATTEMPTS):
         if len(self.entities) >= self.config.NENT:
            break

         r, c   = self.config.SPAWN()
         if self.realm.map.tiles[r, c].occupied:
            continue

         self.spawnIndividual(r, c)

      while len(self.entities) == 0:
         self.spawn()

class Realm:
   '''Top-level world object'''
   def __init__(self, config, identify: Callable):
      self.config   = config
      self.identify = identify

      #Load the world file
      self.dataframe = trinity.Dataframe(config)
      self.map       = core.Map(config, self)

      #Entity handlers
      self.players  = PlayerManager(config, self, identify)
      self.npcs     = NPCManager(config, self)

   def reset(self, idx):
      '''Reset the environment and load the specified map

      Args:
         idx: Map index to load
      ''' 
      self.map.reset(self, idx)
      self.players.reset()
      self.npcs.reset()
      self.tick = 0
 
   def packet(self):
      '''Client packet'''
      return {'environment': self.map.repr,
              'resource': self.map.packet,
              'player': self.players.packet,
              'npc': self.npcs.packet}

   @property
   def population(self):
      '''Number of player agents'''
      return len(self.players.entities)

   def entity(self, entID):
      '''Get entity by ID'''
      if entID < 0:
         return self.npcs[entID]
      else:
         return self.players[entID]

   def step(self, actions):
      '''Run game logic for one tick
      
      Args:
         actions: Dict of agent actions
      '''
      #Prioritize actions
      npcActions = self.npcs.actions(self)
      merged     = defaultdict(list)
      prioritized(actions, merged)
      prioritized(npcActions, merged)

      #Update entities and perform actions
      self.players.update(actions)
      self.npcs.update(npcActions)

      #Execute actions
      for priority in sorted(merged):
         for entID, (atn, args) in merged[priority]:
            ent = self.entity(entID)
            atn.call(self, ent, *args)

      #Cull dead agents and spawn new ones
      dead = self.players.cull()
      self.npcs.cull()

      self.players.spawn()
      self.npcs.spawn()

      #Update map
      self.map.step()
      self.tick += 1

      return dead



from pdb import set_trace as T
import numpy as np
import random as rand

from neural_mmo.forge.blade.systems.ai import utils
from neural_mmo.forge.blade.io.action import static as Action

def random():
   return rand.choice(Action.Direction.edges)

def randomSafe(tiles, ent):
   r, c  = ent.base.pos
   cands = []
   if not tiles[r-1, c].lava:
      cands.append(Action.North)
   if not tiles[r+1, c].lava:
      cands.append(Action.South)
   if not tiles[r, c-1].lava:
      cands.append(Action.West)
   if not tiles[r, c+1].lava:
      cands.append(Action.East)
   
   return rand.choice(cands)

def habitable(tiles, ent):
   r, c  = ent.base.pos
   cands = []
   if tiles[r-1, c].vacant:
      cands.append(Action.North)
   if tiles[r+1, c].vacant:
      cands.append(Action.South)
   if tiles[r, c-1].vacant:
      cands.append(Action.West)
   if tiles[r, c+1].vacant:
      cands.append(Action.East)
   
   if len(cands) == 0:
      return Action.North

   return rand.choice(cands)

def towards(direction):
   if direction == (-1, 0):
      return Action.North
   elif direction == (1, 0):
      return Action.South
   elif direction == (0, -1):
      return Action.West
   elif direction == (0, 1):
      return Action.East
   else:
      return random()

def bullrush(ent, targ):
   direction = utils.directionTowards(ent, targ)
   return towards(direction)

def pathfind(tiles, ent, targ):
   direction = utils.aStar(tiles, ent.pos, targ.pos)
   return towards(direction)

def antipathfind(tiles, ent, targ):
   er, ec = ent.pos
   tr, tc = targ.pos
   goal   = (2*er - tr , 2*ec-tc)
   direction = utils.aStar(tiles, ent.pos, goal)
   return towards(direction)

  





from pdb import set_trace as T
import numpy as np

import random
from neural_mmo.forge.blade.entity import Player
from neural_mmo.forge.blade.entity import entity
from neural_mmo.forge.blade.systems import combat, equipment, ai, combat, skill
from neural_mmo.forge.blade.lib.enums import Neon
from neural_mmo.forge.blade.io.action import static as Action
from neural_mmo.forge.blade.io.stimulus import Static

class NPC(entity.Entity):
   def __init__(self, realm, pos, iden, name, color, pop):
      super().__init__(realm, pos, iden, name, color, pop)
      self.skills = skill.Combat(self)

   def update(self, realm, actions):
      super().update(realm, actions)

      if not self.alive:
         return

      self.resources.health.increment(1)
      self.lastAction = actions

   @staticmethod
   def spawn(realm, pos, iden):
      config = realm.config

      #Select AI Policy
      danger = combat.danger(config, pos)
      if danger >= config.NPC_SPAWN_AGGRESSIVE:
         ent = Aggressive(realm, pos, iden)
      elif danger >= config.NPC_SPAWN_NEUTRAL:
         ent = PassiveAggressive(realm, pos, iden)
      elif danger >= config.NPC_SPAWN_PASSIVE:
         ent = Passive(realm, pos, iden)
      else:
         return

      #Set levels
      levels = NPC.clippedLevels(config, danger, n=5)
      constitution, defense, melee, ranged, mage = levels

      ent.resources.health.max = constitution
      ent.resources.health.update(constitution)

      ent.skills.constitution.setExpByLevel(constitution)
      ent.skills.defense.setExpByLevel(defense)
      ent.skills.melee.setExpByLevel(melee)
      ent.skills.range.setExpByLevel(ranged)
      ent.skills.mage.setExpByLevel(mage)

      ent.skills.style = random.choice(
         (Action.Melee, Action.Range, Action.Mage))

      #Set equipment levels
      ent.loadout.chestplate.level = NPC.gearLevel(defense)
      ent.loadout.platelegs.level  = NPC.gearLevel(defense)

      return ent

   def yieldDrops(self):
      self.lastAttacker.receiveDrops(self.drops.roll())

   @staticmethod
   def gearLevel(lvl, offset=10):
      proposed = random.gauss(lvl-offset, offset)
      lvl      = np.clip(proposed, 0, lvl)
      return int(lvl)

   @staticmethod
   def clippedLevels(config, danger, n=1):
      lmin    = config.NPC_LEVEL_MIN
      lmax    = config.NPC_LEVEL_MAX

      lbase   = danger*(lmax-lmin) + lmin
      lspread = config.NPC_LEVEL_SPREAD

      lvlMin  = int(max(lmin, lbase - lspread))
      lvlMax  = int(min(lmax, lbase + lspread))

      lvls = [random.randint(lvlMin, lvlMax) for _ in range(n)]

      if n == 1:
         return lvls[0]

      return lvls
 
   def packet(self):
      data = super().packet()

      data['base']     = self.base.packet()      
      data['skills']   = self.skills.packet()      
      data['resource'] = {'health': self.resources.health.packet()}

      return data

   @property
   def isNPC(self) -> bool:
      return True

class Passive(NPC):
   def __init__(self, realm, pos, iden):
      super().__init__(realm, pos, iden, 'Passive', Neon.GREEN, -1)
      self.dataframe.init(Static.Entity, iden, pos)

   def decide(self, realm):
      return ai.policy.passive(realm, self)

class PassiveAggressive(NPC):
   def __init__(self, realm, pos, iden):
      super().__init__(realm, pos, iden, 'Neutral', Neon.ORANGE, -2)
      self.dataframe.init(Static.Entity, iden, pos)

   def decide(self, realm):
      return ai.policy.neutral(realm, self)

class Aggressive(NPC):
   def __init__(self, realm, pos, iden):
      super().__init__(realm, pos, iden, 'Hostile', Neon.RED, -3)
      self.dataframe.init(Static.Entity, iden, pos)
      self.vision = int(max(self.vision, 1 + combat.level(self.skills) // 10))
      self.dataframe.init(Static.Entity, self.entID, self.pos)

   def decide(self, realm):
      return ai.policy.hostile(realm, self)



from neural_mmo.forge.blade.item import item, armor

class Inventory:
   def __init__(self, config):
      self.ammo = 0
      #self.resetArmor()
      #self.resetMelee()
      #self.resetRanged()

   def resetArmor(self):
      self.armor = Armor.Base()

   def resetMelee(self):
      self.melee = Item.Base()
      
   def resetRanged(self):
      self.ranged = Item.Base()



from pdb import set_trace as T

import numpy as np


SEP     = u'\u2595\u258f'
BLOCK   = u'\u2591'
TOP     = u'\u2581'
BOT     = u'\u2594'
LEFT    = u'\u258f'
RIGHT   = u'\u2595'

def table_stats(stats, titleLen=12, entryLen=12):
   titleFmt = '{:<' + str(titleLen) + '}'
   valFmt   = '{:' + str(entryLen) + '.2f}'
   keyFmt   = '{:<' + str(entryLen) + '}'

   keys  = [keyFmt.format(k) for k in 'Min Max Mean Std'.split()]
   title = titleFmt.format('Metric')
   lines = [[title] + keys]

   for key, stat in stats.items():
      l = [titleFmt.format(key)]
      for func in (np.min, np.max, np.mean, np.std):
         val = valFmt.format(func(stat))
         l.append(val)
      lines.append(l)
      
   llens = [titleLen] + 4*[entryLen]
   seps  = ['='*l for l in llens]
   lines = [seps, lines[0], seps, *(lines[1:]), seps]
   lines = [' '.join(e) for e in lines]
 
   return lines


def precomputed_stats(stats):
   '''Format a dict of precomputed stats'''
   lines = []
   for key, stat in stats.items():
      keys  = 'Min Max Mean Std'.split()
      vals  = [stat[k] for k in keys]
      lines.append(line(title=key, keys=keys, vals=vals))
   
   return lines

def stats(stats):
   '''Format a dict of stats'''
   lines = []
   for key, stat in stats.items():
      mmin, mmax = np.min(stat),  np.max(stat)
      mean, std  = np.mean(stat), np.std(stat)

      lines.append(line(
            title = key,
            keys  = 'Min Max Mean Std'.split(),
            vals  = [mmin, mmax, mean, std]))

   return lines

def times(stats):
   '''Format a dict of timed data'''
   lines = []
   for key, stat in stats.items():
      ssum, n   = np.sum(stat),  len(stat)
      mean, std = np.mean(stat), np.std(stat)

      lines.append(line(
            title = key,
            keys  = 'Total N Mean Std'.split(),
            vals  = [ssum, n, mean, std]))

   return lines

def line(title=None, keys=[], vals=[],
      titleFmt='{:<12}', keyFmt='{}', valFmt='{:8.1f}'):
   '''Format a line of stats with vertical separators'''

   assert len(keys) == len(vals), 'Unequal number of keys and values'

   fmt = []
   if title is not None:
      fmt.append(titleFmt.format(title))

   for key, val in zip(keys, vals):
      fmt.append((keyFmt + ': ' + valFmt).format(key, val))

   return SEP.join(fmt)

def box(block, indent=1):
   '''Indent lines and draw a box around them'''
   mmax   = max(len(line) for line in block) + 2
   fmt    = '{:<'+str(mmax+1)+'}'
   indent = ' ' * indent

   for idx, l in enumerate(block):
      block[idx] = indent + fmt.format(LEFT + l + RIGHT)

   block.insert(0, indent + TOP*mmax)
   block.append(indent + BOT*mmax)
   return block

def box_stats(vals, indent=1):
   '''Format stats and then draw a box around them'''
   return box(stats(vals))
   



from pdb import set_trace as T
import numpy as np
import gym

import inspect

from neural_mmo.forge.blade.lib.utils import classproperty
from neural_mmo.forge.blade.io.comparable import IterableTypeCompare

class Flat:
   pass

class Stim(metaclass=IterableTypeCompare):
   default = 0
   max = np.inf
   min = 0

   def __init__(self, config):
      cls = self.__class__
      self.min = cls.min
      self.max = cls.max
      self.default = cls.default

      self.init(config)
      self._val = self.default

   def init(self, config):
      pass

   @classproperty
   def name(self):
      name = self.__name__
      return name[0].lower() + name[1:]

   @property
   def val(self):
      if self._val is None:
         return 0
      return self._val

   def asserts(self, val):
      if val is not None:
         assert val >= self.min and val <= self.max, str(self) + ': ' + str(val)
      return val
 
   def update(self, val):
      self._val = val
      return self #for convenience

   def packet(self):
      return {
            'val': self.val, 
            'max': self.max if self.max != float('inf') else None}

   def get(self, *args):
      return self.asserts(self.val)

   @property
   def missing(self):
      return self.max - self.val

   def increment(self, amt=1):
      self._val = min(self.max, self.val + amt)

   def decrement(self, amt=1):
      self._val = max(0, self.val - amt)

   def __add__(self, other):
      self.increment(other)
      return self

   def __sub__(self, other):
      self.decrement(other)
      return self

   def __lt__(self, other):
      return self.val < other

   def __le__(self, other):
      return self.val <= other

   def __gt__(self, other):
      return self.val > other

   def __ge__(self, other):
      return self.val >= other

class Discrete(Stim):
   def __init__(self, config):
      super().__init__(config)      
      self.space = gym.spaces.Box(
            low=np.float32(0.0),
            high=np.float32(self.range),
            shape=(1,))

   @property
   def range(self):
      return self.max - self.min + 1

   def oneHot(self):
      ary = np.zeros(self.range)
      ary[self.norm()] = 1
      return ary

   def norm(self):
      val = self.val# - self.min
      assert val == int(val)
      return int(val)

   def get(self, *args):
      self.asserts(self.val)
      return np.array([self.norm()])
      return self.norm()

      #No norm needed for discrete vars. Below is for
      #current hack where RLlib treats everything as continuous
      #The default preprocessor won't norm, so we can still embed
      return self.norm()

class Continuous(Stim):
   def __init__(self, config):
      super().__init__(config)      
      self.space = gym.spaces.Box(
            low=np.float32(-1.0),
            high=np.float32(1.0),
            shape=(1,))

   @property
   def range(self):
      return self.max - self.min

   def norm(self):
      assert self.val >= self.min and self.val <= self.max
      val = self.val - self.min
      if self.range == np.inf:
         val = self.scaled(val)
      else:
         val = 2*(val / self.range) - 1
      assert val >= -1 and val <= 1, self
      return val

   def scaled(self, val):
      return self.scale * val

   def get(self, *args):
      self.asserts(self.val)
      val = self.norm()
      return np.array([val])




from pdb import set_trace as T
import numpy as np

from collections import defaultdict
from itertools import chain
from copy import deepcopy

from neural_mmo.forge.blade import entity, core
from neural_mmo.forge.blade.io import stimulus
from neural_mmo.forge.blade.io.stimulus import Static
from neural_mmo.forge.blade.systems import combat

from neural_mmo.forge.blade.lib.enums import Palette
from neural_mmo.forge.blade.lib import log

class Env:
   '''Environment wrapper for Neural MMO

   Note that the contents of (ob, reward, done, info) returned by the standard
   OpenAI Gym API reset() and step(actions) methods have been generalized to
   support variable agent populations and more expressive observation/action
   spaces. This means you cannot use preexisting optimizer implementations that
   strictly expect the OpenAI Gym API. We recommend PyTorch+RLlib to take
   advantage of our prebuilt baseline implementations, but any framework that
   supports RLlib's fairly popular environment API and extended OpenAI
   gym.spaces observation/action definitions should work as well.'''
   def __init__(self, config):
      '''
      Args:
         config : A forge.blade.core.Config object or subclass object
      '''
      super().__init__()
      self.realm     = core.Realm(config, self.spawn)

      self.config    = config
      self.overlay   = None

      self.dead      = []
      self.spawned   = 0

      #self.steps = 0

   ############################################################################
   ### Core API
   def reset(self, idx=None, step=True):
      '''OpenAI Gym API reset function

      Loads a new game map and returns initial observations

      Args:
         idx: Map index to load. Selects a random map by default

         step: Whether to step the environment and return initial obs

      Returns:
         obs: Initial obs if step=True, None otherwise 

      Notes:
         Neural MMO simulates a persistent world. Ideally, you should reset
         the environment only once, upon creation. In practice, this approach
         limits the number of parallel environment simulations to the number
         of CPU cores available. At small and medium hardware scale, we
         therefore recommend the standard approach of resetting after a long
         but finite horizon: ~1000 timesteps for small maps and
         5000+ timesteps for large maps

      Returns:
         observations, as documented by step()
      '''
      self.quill = log.Quill(self.realm.identify)
      
      if idx is None:
         idx = np.random.randint(self.config.TERRAIN_TRAIN_MAPS) + 1

      self.worldIdx = idx
      self.realm.reset(idx)

      obs = None
      if step:
         obs, _, _, _ = self.step({})

      return obs

   def step(self, actions, preprocess=set(), omitDead=True):
      '''OpenAI Gym API step function simulating one game tick or timestep

      Args:
         actions: A dictionary of agent decisions of format::

               {
                  agent_1: {
                     action_1: [arg_1, arg_2],
                     action_2: [...],
                     ...
                  },
                  agent_2: {
                     ...
                  },
                  ...
               }

            Where agent_i is the integer index of the i\'th agent 

            The environment only evaluates provided actions for provided
            agents. Unprovided action types are interpreted as no-ops and
            illegal actions are ignored

            It is also possible to specify invalid combinations of valid
            actions, such as two movements or two attacks. In this case,
            one will be selected arbitrarily from each incompatible sets.

            A well-formed algorithm should do none of the above. We only
            Perform this conditional processing to make batched action
            computation easier.

         preprocess: set of agent IDs for which actions are returned as raw
            indices and need to be preprocessed. Typically this should only
            include IDs of agents controlled by neural models and exclude IDs
            of scripted agents

         omitDead: Whether to omit dead agents observations from the returned
            obs. Provided for conformity with some optimizer APIs

      Returns:
         (dict, dict, dict, None):

         observations:
            A dictionary of agent observations of format::

               {
                  agent_1: obs_1,
                  agent_2: obs_2,
                  ...
               ]

            Where agent_i is the integer index of the i\'th agent and
            obs_i is the observation of the i\'th' agent. Note that obs_i
            is a structured datatype -- not a flat tensor. It is automatically
            interpretable under an extended OpenAI gym.spaces API. Our demo
            code shows how do to this in RLlib. Other frameworks must
            implement the same extended gym.spaces API to do the same.
            
         rewards:
            A dictionary of agent rewards of format::

               {
                  agent_1: reward_1,
                  agent_2: reward_2,
                  ...
               ]

            Where agent_i is the integer index of the i\'th agent and
            reward_i is the reward of the i\'th' agent.

            By default, agents receive -1 reward for dying and 0 reward for
            all other circumstances. Override Env.reward to specify
            custom reward functions
 
         dones:
            A dictionary of agent done booleans of format::

               {
                  agent_1: done_1,
                  agent_2: done_2,
                  ...
               ]

            Where agent_i is the integer index of the i\'th agent and
            done_i is a boolean denoting whether the i\'th agent has died.

            Note that obs_i will be a garbage placeholder if done_i is true.
            This is provided only for conformity with OpenAI Gym. Your
            algorithm should not attempt to leverage observations outside of
            trajectory bounds. You can omit garbage obs_i values by setting
            omitDead=True.

         infos:
            An empty dictionary provided only for conformity with OpenAI Gym.
      '''
      #Preprocess actions
      for entID in preprocess:
         ent = self.realm.players[entID]
         if not ent.alive:
            continue

         for atn, args in actions[entID].items():
            for arg, val in args.items():
               if len(arg.edges) > 0:
                  actions[entID][atn][arg] = arg.edges[val]
               elif val < len(ent.targets):
                  targ                     = ent.targets[val]
                  actions[entID][atn][arg] = self.realm.entity(targ)
               else: #Need to fix -inf in classifier before removing this
                  actions[entID][atn][arg] = ent

      #Step: Realm, Observations, Logs
      self.dead = self.realm.step(actions)
      obs, rewards, dones, self.raw = {}, {}, {}, {}
      for entID, ent in self.realm.players.items():
         ob             = self.realm.dataframe.get(ent)
         obs[entID]     = ob
         self.dummy_ob  = ob

         rewards[entID] = self.reward(ent)
         dones[entID]   = False

      #self.steps += len(self.realm.players.items())
      #print('World {} Tick {} Steps {}'.format(self.worldIdx, self.realm.tick, self.steps))

      for entID, ent in self.dead.items():
         self.log(ent)

      #Postprocess dead agents
      if omitDead:
         return obs, rewards, dones, {}

      for entID, ent in self.dead.items():
         rewards[ent.entID] = self.reward(ent)
         dones[ent.entID]   = True
         obs[ent.entID]     = self.dummy_ob

      return obs, rewards, dones, {}

   ############################################################################
   ### Logging
   def log(self, ent) -> None:
      '''Logs agent data upon death

      This function is called automatically when an agent dies. Logs are used
      to compute summary stats and populate the dashboard. You should not
      call it manually. Instead, override this method to customize logging.

      Args:
         ent: An agent
      '''

      quill = self.quill

      blob = quill.register('Population', self.realm.tick,
            quill.HISTOGRAM, quill.LINE, quill.SCATTER)
      blob.log(self.realm.population)

      blob = quill.register('Lifetime', self.realm.tick,
            quill.HISTOGRAM, quill.LINE, quill.SCATTER, quill.GANTT)
      blob.log(ent.history.timeAlive.val)

      blob = quill.register('Skill Level', self.realm.tick,
            quill.HISTOGRAM, quill.STACKED_AREA, quill.STATS, quill.RADAR)
      blob.log(ent.skills.range.level,        'Range')
      blob.log(ent.skills.mage.level,         'Mage')
      blob.log(ent.skills.melee.level,        'Melee')
      blob.log(ent.skills.constitution.level, 'Constitution')
      blob.log(ent.skills.defense.level,      'Defense')
      blob.log(ent.skills.fishing.level,      'Fishing')
      blob.log(ent.skills.hunting.level,      'Hunting')

      blob = quill.register('Equipment', self.realm.tick,
            quill.HISTOGRAM, quill.SCATTER)
      blob.log(ent.loadout.chestplate.level, 'Chestplate')
      blob.log(ent.loadout.platelegs.level,  'Platelegs')

      blob = quill.register('Exploration', self.realm.tick,
            quill.HISTOGRAM, quill.SCATTER)
      blob.log(ent.history.exploration)

      quill.stat('Lifetime',  ent.history.timeAlive.val)

      if self.config.game_system_enabled('Achievement'):
         quill.stat('Achievement', ent.achievements.score())
         for name, stat in ent.achievements.stats:
            quill.stat(name, stat)

   def terminal(self):
      '''Logs currently alive agents and returns all collected logs

      Automatic log calls occur only when agents die. To evaluate agent
      performance over a fixed horizon, you will need to include logs for
      agents that are still alive at the end of that horizon. This function
      performs that logging and returns the associated a data structure
      containing logs for the entire evaluation

      Returns:
         Log datastructure. Use them to update an InkWell logger.
         
      Args:
         ent: An agent
      '''

      for entID, ent in self.realm.players.entities.items():
         self.log(ent)

      return self.quill.packet

   ############################################################################
   ### Override hooks
   def reward(self, entID):
      '''Computes the reward for the specified agent

      Override this method to create custom reward functions. You have full
      access to the environment state via self.realm. Our baselines do not
      modify this method; specify any changes when comparing to baselines

      Returns:
         float:

         reward:
            The reward for the actions on the previous timestep of the
            entity identified by entID.
      '''
      if entID not in self.realm.players:
         return -1
      return 0

   def spawn(self):
      '''Called when an agent is added to the environment

      Override this method to specify name/population upon spawning.

      Returns:
         (int, str):

         popID:
            An integer used to identity membership within a population

         prefix:
            The agent will be named prefix + entID

      Notes:
         Mainly intended for population-based research. In particular, it
         allows you to define behavior that selectively spawns agents into
         particular populations based on the current game state -- for example,
         current population sizes or performance.'''

      popSize = self.config.NENT / self.config.NPOP
      pop     = (self.spawned // popSize) % self.config.NPOP
      self.spawned += 1
      return int(pop), 'Neural_'

   ############################################################################
   ### Client data
   @property
   def packet(self):
      '''Data packet used by the renderer

      Returns:
         packet: A packet of data for the client
      '''
      packet = {
            'config': self.config,
            'pos': self.overlayPos,
            'wilderness': 0
            }

      packet = {**self.realm.packet(), **packet}

      if self.overlay is not None:
         print('Overlay data: ', len(self.overlay))
         packet['overlay'] = self.overlay
         self.overlay      = None

      return packet

   def register(self, overlay):
      '''Register an overlay to be sent to the client

      The intended use of this function is: User types overlay ->
      client sends cmd to server -> server computes overlay update -> 
      register(overlay) -> overlay is sent to client -> overlay rendered

      Args:
         values: A map-sized (self.size) array of floating point values
      '''
      err = 'overlay must be a numpy array of dimension (*(env.size), 3)'
      assert type(overlay) == np.ndarray, err
      self.overlay = overlay.tolist()

   def dense(self):
      '''Simulates an agent on every tile and returns observations

      This method is used to compute per-tile visualizations across the
      entire map simultaneously. To do so, we spawn agents on each tile
      one at a time. We compute the observation for each agent, delete that
      agent, and go on to the next one. In this fashion, each agent receives
      an observation where it is the only agent alive. This allows us to
      isolate potential influences from observations of nearby agents

      This function is slow, and anything you do with it is probably slower.
      As a concrete example, consider that we would like to visualize a
      learned agent value function for the entire map. This would require
      computing a forward pass for one agent per tile. To cut down on
      computation costs, we omit lava tiles from this method

      Returns:
         (dict, dict):

         observations:
            A dictionary of agent observations as specified by step()

         ents:
            A corresponding dictionary of agents keyed by their entID
      '''
      config  = self.config
      R, C    = self.realm.map.tiles.shape

      entID   = 100000
      pop     = 0
      name    = "Value"
      color   = (255, 255, 255)


      observations, ents = {}, {}
      for r in range(R):
         for c in range(C):
            tile    = self.realm.map.tiles[r, c]
            if not tile.habitable:
               continue

            current = tile.ents
            n       = len(current)
            if n == 0:
               ent = entity.Player(self.realm, (r, c), entID, pop, name, color)
            else:
               ent = list(current.values())[0]

            obs = self.realm.dataframe.get(ent)
            if n == 0:
               self.realm.dataframe.remove(Static.Entity, entID, ent.pos)

            observations[entID] = obs
            ents[entID] = ent
            entID += 1

      return observations, ents



from pdb import set_trace as T
import numpy as np

from neural_mmo.forge.blade.io.stimulus.static import Stimulus
from neural_mmo.forge.blade.io.action import static as Action

from neural_mmo.forge.trinity.scripted import io, utils

def closestTarget(config, ob):
   shortestDist = np.inf
   closestAgent = None

   Entity = Stimulus.Entity
   agent  = ob.agent

   sr = io.Observation.attribute(agent, Entity.R)
   sc = io.Observation.attribute(agent, Entity.C)
   start = (sr, sc)

   for target in ob.agents:
      exists = io.Observation.attribute(target, Entity.Self)
      if not exists:
         continue

      tr = io.Observation.attribute(target, Entity.R)
      tc = io.Observation.attribute(target, Entity.C)

      goal = (tr, tc)
      dist = utils.l1(start, goal)

      if dist < shortestDist and dist != 0:
          shortestDist = dist
          closestAgent = target

   if closestAgent is None:
      return None, None

   return closestAgent, shortestDist

def attacker(config, ob):
   Entity = Stimulus.Entity

   sr = io.Observation.attribute(ob.agent, Entity.R)
   sc = io.Observation.attribute(ob.agent, Entity.C)
 
   attackerID = io.Observation.attribute(ob.agent, Entity.AttackerID)

   if attackerID == 0:
       return None, None

   for target in ob.agents:
      identity = io.Observation.attribute(target, Entity.ID)
      if identity == attackerID:
         tr = io.Observation.attribute(target, Entity.R)
         tc = io.Observation.attribute(target, Entity.C)
         dist = utils.l1((sr, sc), (tr, tc))
         return target, dist
   return None, None

def target(config, actions, style, targetID):
   actions[Action.Attack] = {
         Action.Style: style,
         Action.Target: targetID}




import inspect

from neural_mmo.forge.blade.io.stimulus import node


class IterableTypeCompare(type):
   def __iter__(cls):
      stack = list(cls.__dict__.items())
      while len(stack) > 0:
         name, attr = stack.pop()
         if type(name) != tuple:
            name = tuple([name])
         if not inspect.isclass(attr):
            continue
         if issubclass(attr, node.Flat):
            for n, a in attr.__dict__.items():
               n = name + tuple([n]) 
               stack.append((n, a))
            continue
         yield name, attr

   def values(cls):
      return [e[1] for e in cls]

   def __hash__(self):
      return hash(self.__name__)

   def __eq__(self, other):
      return self.__name__ == other.__name__

   def __ne__(self, other):
      return self.__name__ != other.__name__

   def __lt__(self, other):
      return self.__name__ < other.__name__

   def __le__(self, other):
      return self.__name__ <= other.__name__

   def __gt__(self, other):
      return self.__name__ > other.__name__

   def __ge__(self, other):
      return self.__name__ >= other.__name__

class Iterable(type):
   def __iter__(cls):
      stack = list(cls.__dict__.items())
      while len(stack) > 0:
         name, attr = stack.pop()
         if name.startswith('__'):
            continue
         yield name, attr





from neural_mmo.forge.blade.core.map import Map
from neural_mmo.forge.blade.core.realm import Realm
from neural_mmo.forge.blade.core.tile import Tile
from neural_mmo.forge.blade.core.config import Config



from pdb import set_trace as T
from neural_mmo.forge.blade.lib.enums import Tier

class Loadout:
   def __init__(self, chest=0, legs=0):
      self.chestplate = Chestplate(chest)
      self.platelegs  = Platelegs(legs) 

   @property
   def defense(self):
      return (self.chestplate.level + self.platelegs.level) // 2

   def packet(self):
      packet = {}

      packet['chestplate'] = self.chestplate.packet()
      packet['platelegs']  = self.platelegs.packet()

      return packet

class Armor:
   def __init__(self, level):
      self.level = level

   def packet(self):
     packet = {}
 
     packet['level'] = self.level
     packet['color'] = self.color.packet()

     return packet

   @property
   def color(self):
     if self.level == 0:
        return Tier.BLACK
     if self.level < 10:
        return Tier.WOOD
     elif self.level < 20:
        return Tier.BRONZE
     elif self.level < 40:
        return Tier.SILVER
     elif self.level < 60:
        return Tier.GOLD
     elif self.level < 80:
        return Tier.PLATINUM
     else:
        return Tier.DIAMOND
     

class Chestplate(Armor):
   pass

class Platelegs(Armor):
   pass




from pdb import set_trace as T
import numpy as np

import torch
from torch import nn

from neural_mmo.forge.ethyr.torch.policy import linear, functional

class DecomposedAttention(nn.Module):
   def __init__(self, h):
      super().__init__()
      self.scale = h

   def forward(self, X, Q, K, V):
      X   = Q.transpose(-2, -1)
      QX  = torch.matmul(Q, X)
      KXT = torch.matmul(K, X).transpose(-2, -1)
      VX  = torch.matmul(V, X) / self.scale 
      
      KXTVX   = torch.matmul(KXT, VX)
      QXKXTVX = torch.matmul(QX, KXTVX)

      return QXKXTVX

class ScaledDotProductAttention(nn.Module):
   def __init__(self, h):
      super().__init__()
      self.scale = np.sqrt(h)

   def forward(self, Q, K, V):
      Kt  = K.transpose(-2, -1)
      QK  = torch.matmul(Q, Kt)

      #Original is attending over hidden dims?
      #QK  = torch.softmax(QK / self.scale, dim=-2)
      QK    = torch.softmax(QK / self.scale, dim=-1)
      score = torch.sum(QK, -2)
      QKV = torch.matmul(QK, V)
      return QKV, score

class SelfAttention(nn.Module):
   def __init__(self, xDim, yDim, flat=True):
      super().__init__()

      self.Q = torch.nn.Linear(xDim, yDim)
      self.K = torch.nn.Linear(xDim, yDim)
      self.V = torch.nn.Linear(xDim, yDim)

      self.attention = ScaledDotProductAttention(yDim)
      self.flat = flat

   def forward(self, q):
      Q = self.Q(q)
      K = self.K(q)
      V = self.V(q)

      attn, scores = self.attention(Q, K, V)

      if self.flat:
         attn, _ = torch.max(attn, dim=-2)

      return attn, scores

class Attention(nn.Module):
   def __init__(self, xDim, yDim, flat=True):
      super().__init__()

      self.Q = torch.nn.Linear(xDim, yDim)
      self.K = torch.nn.Linear(xDim, yDim)
      self.V = torch.nn.Linear(xDim, yDim)

      self.attention = ScaledDotProductAttention(yDim)
      self.flat = flat

   def forward(self, q, v):
      Q = self.Q(q)
      K = self.K(v)
      V = self.V(v)

      #??? Doing 100x100 mat here, not good
      attn, scores = self.attention(Q, K, V)

      if self.flat:
         attn, _ = torch.max(attn, dim=-2)

      return attn, scores

class MaxReluBlock(nn.Module):
   def __init__(self, h, layers=2):
      super().__init__()
      self.net = linear.ReluBlock(
         h, layers, postRelu=False)
 
   def forward(self, x):
      x    = self.net(x)
      x, _ = torch.max(x, dim=-2)
      return x

class DotReluBlock(nn.Module):
   def __init__(self, h, layers=2):
      super().__init__()
      self.key = linear.ReluBlock(
         h, layers, postRelu=False)

      self.val = linear.ReluBlock(
         h, layers, postRelu=False)
   
   def forward(self, k, v):
      k = self.key(k).unsqueeze(-2)
      v = self.val(v)
      x = torch.sum(k * v, -1)
      return x



'''Action decision module'''

from pdb import set_trace as T
import numpy as np

from collections import defaultdict

import torch
from torch import nn

from neural_mmo.forge.blade.io.action import static
from neural_mmo.forge.blade.io.stimulus.static import Stimulus

from neural_mmo.forge.ethyr.torch.policy import attention
from neural_mmo.forge.ethyr.torch.policy import functional

class Output(nn.Module):
   def __init__(self, config):
      '''Network responsible for selecting actions

      Args:
         config: A Config object
      '''
      super().__init__()
      self.config = config
      self.h = config.HIDDEN

      self.net = DiscreteAction(self.config, self.h, self.h)
      self.arg = nn.Embedding(static.Action.n, self.h)
      #self.net = FlatAction(self.config, self.h, self.h)

   def names(self, nameMap, args):
      '''Lookup argument indices from name mapping'''
      return np.array([nameMap.get(e) for e in args])

   def forward(self, obs, lookup):
      '''Populates an IO object with actions in-place                         
                                                                              
      Args:                                                                   
         obs               : An IO object specifying observations
         vals              : A value prediction for each agent
         observationTensor : A fixed size observation representation
         entityLookup      : A fixed size representation of each entity
         manager           : A RolloutManager object
      ''' 
      rets = defaultdict(dict)
      for atn in static.Action.edges:
         for arg in atn.edges:
            lens  = None
            if arg.argType == static.Fixed:
               batch = obs.shape[0]
               idxs  = [e.idx for e in arg.edges]
               #cands = lookup[static.Fixed.__name__][idxs]
               cands = self.arg.weight[idxs]
               cands = cands.repeat(batch, 1, 1)
               #Fixed arg
            else:
               #Temp hack, rename
               cands = lookup['Entity']
               lens  = lookup['N']

            logits = self.net(obs, cands, lens)

            #String names for RLlib for now
            #rets[atn.__name__][arg.__name__] = logits
            rets[atn][arg] = logits

      return rets
      
#Root action class
class Action(nn.Module):
   pass

class DiscreteAction(Action):
   '''Head for making a discrete selection from
   a variable number of candidate actions'''
   def __init__(self, config, xdim, h):
      super().__init__()
      self.net = attention.DotReluBlock(h)

   def forward(self, stim, args, lens):
      x = self.net(stim, args)

      if lens is not None:
         mask = torch.arange(x.shape[-1]).to(x.device).expand_as(x)
         x[mask >= lens] = 0

      return x



from pdb import set_trace as T
import numpy as np

class ExperienceCalculator:
   def __init__(self):
      self.exp = [0]
      self.tabulateExp()

   def tabulateExp(self, numLevels=99):
      for i in range(2, numLevels+1):
         increment = np.floor(i-1 + 300*(2**((i-1)/7.0)))/4.0
         self.exp += [self.exp[-1] + increment]

      self.exp = np.floor(np.array(self.exp))

   def expAtLevel(self, level):
      return self.exp[level - 1]

   def levelAtExp(self, exp):
      if exp >= self.exp[-1]:
         return len(self.exp)
      return np.argmin(exp >= self.exp)



import heapq, itertools
import itertools

class PriorityQueue:
   def __init__(self, capacity, unique=False):
      self.q, self.items = [], set()
      self.capacity = capacity
      self.count = itertools.count()
      self.unique = unique

   def get(self, ind):
      priority, item = self.tolist()[ind]
      return priority, item

   def push(self, item, priority, uniqueKey=None):
      if self.unique:
         self.items.add(uniqueKey)
      count = next(self.count)
      if len(self.q) >= self.capacity:
         return heapq.heappushpop(self.q, (priority, count, item))
      heapq.heappush(self.q, (priority, count, item))

   def pop(self):
      priority, _, item = heapq.heappop(self.q)
      if self.unique:
         self.items.remove(item)
      return priority, item

   @property
   def peek(self):
      return self.peekPriority, self.peekValue

   @property
   def peekPriority(self):
      ret = heapq.nlargest(1, self.q)
      if len(ret) > 0:
         return ret[0][0]

   @property
   def peekValue(self):
      ret = heapq.nlargest(1, self.q)
      if len(ret) > 0:
         return ret[0][2]


   def tolist(self):
      q = heapq.nlargest(self.n, self.q)
      return [(e[0], e[2]) for e in q]

   def priorities(self):
      return sorted([e[0] for e in self.q], reverse=True)

   def print(self):
      q = heapq.nlargest(self.n, self.q)
      print([(e[0]) for e in q], end='')
      print()

   @property
   def n(self):
      return len(self.q)




import numpy as np

class Range():
   def __init__(self, mmin, mmax):
      self.mmin = mmin
      self.mmax = mmax

   @property
   def value(self):
      return np.random.randint(self.mmin, self.mmax+1)

class Drop:
   def __init__(self, item, amount, prob):
      self.item = item
      self.amount = amount
      self.prob = prob

   def roll(self):
      if np.random.rand() < self.prob:
         if type(self.amount) == int:
            return (self.item(), self.amount)
         return (self.item(), self.amount.value)

class DropTable:
   def __init__(self):
      self.drops = []

   def add(self, item, quant, prob=1.0):
      self.drops += [Drop(item, quant, prob)]

   def roll(self):
      ret = []
      for e in self.drops:
         drop = e.roll()
         if drop is not None:
            ret += [drop]
      return ret



import numpy as np

from collections import defaultdict
from tqdm import tqdm

from neural_mmo.forge.trinity import Env
from neural_mmo.forge.trinity.overlay import OverlayRegistry

from neural_mmo.forge.blade.io.action import static as Action
from neural_mmo.forge.blade.lib.log import InkWell

class Base:
   '''Base class for test-time evaluators'''
   def __init__(self, config):
      self.config  = config
      self.done    = {}

   def render(self):
      '''Rendering launches a Twisted WebSocket server with a fixed
      tick rate. This is a blocking call; the server will handle 
      environment execution using the provided tick function.'''
      from neural_mmo.forge.trinity.twistedserver import Application
      Application(self.env, self.tick)

   def evaluate(self, generalize=True):
      '''Evaluate the model on maps according to config params'''

      config = self.config
      log    = InkWell()

      if generalize:
         maps = range(-1, -config.EVAL_MAPS-1, -1)
      else:
         maps = range(1, config.EVAL_MAPS+1)

      print('Number of evaluation maps: {}'.format(len(maps)))
      for idx in maps:
         self.obs = self.env.reset(idx)
         for t in tqdm(range(config.EVALUATION_HORIZON)):
            self.tick(None, None)

         log.update(self.env.terminal())

      #Save data
      np.save(config.PATH_EVALUATION, log.packet)

   def tick(self, obs, actions, pos, cmd, preprocess=set(), omitDead=True):
      '''Simulate a single timestep

      Args:
          obs: dict of agent observations
          actions: dict of policy actions
          pos: Camera position (r, c) from the server
          cmd: Console command from the server
          preprocessActions: Required for actions provided as indices
      '''
      self.obs, rewards, self.done, _ = self.env.step(
            actions, preprocess, omitDead)
      if self.config.RENDER:
         self.registry.step(obs, pos, cmd)

class Evaluator(Base):
   '''Evaluator for scripted models'''
   def __init__(self, config, policy, *args):
      super().__init__(config)
      self.policies = defaultdict(lambda: policy(config)) 
      self.args     = args

      self.env      = Env(config)

   def render(self):
      '''Render override for scripted models'''
      self.obs      = self.env.reset()
      self.registry = OverlayRegistry(self.config, self.env).init()
      super().render()

   def tick(self, pos, cmd):
      '''Simulate a single timestep

      Args:
          pos: Camera position (r, c) from the server)
          cmd: Console command from the server
      '''
      realm, actions    = self.env.realm, {}
      for agentID in self.obs:
         agent  = realm.players[agentID]
         policy = self.policies[agentID]
         ob     = self.obs[agentID]

         actions[agentID] = policy(ob, *self.args)
         if Action.Attack in actions[agentID]:
            targID = actions[agentID][Action.Attack][Action.Target]
            actions[agentID][Action.Attack][Action.Target] = realm.entity(targID)
         #actions[agentID]   = self.policy(realm, agent, *self.args)

      super().tick(self.obs, actions, pos, cmd)



from pdb import set_trace as T
import numpy as np

import torch
from torch import nn

class BatchFirstLSTM(nn.LSTM):
   def __init__(self, *args, **kwargs):
      super().__init__(*args, batch_first=True, **kwargs)

   def forward(self, input, hx):
      h, c       = hx
      h          = h.transpose(0, 1)
      c          = c.transpose(0, 1)
      hidden, hx = super().forward(input, [h, c])
      h, c       = hx
      h          = h.transpose(0, 1)
      c          = c.transpose(0, 1)
      return hidden, [h, c]
      



from pdb import set_trace as T

import os
import shutil
import zipfile
from pathlib import Path

from setuptools import find_packages, setup
import glob

import versioneer

README = Path("README.md").read_text()
REPO_URL = "https://github.com/jsuarez5341/neural-mmo"
VERSION = versioneer.get_version()

current_dir = Path(__file__).resolve().parent


def read_requirements_file(requirements_version):
    with open(
        str(current_dir / "requirements" / f"{requirements_version}.txt")
    ) as reqs_file:
        reqs = reqs_file.read().split()
    lines_to_remove = []
    for idx in range(len(reqs)):
        if "-r " in reqs[idx]:
            lines_to_remove.append(idx)
    for idx in lines_to_remove:
        reqs.pop(idx)
    return reqs

setup(
    name="neural-mmo",
    description="Neural MMO is a massively multiagent environment for artificial intelligence research inspired by "
    "Massively Multiplayer Online (MMO) role-playing games",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    version=VERSION,
    cmdclass=versioneer.get_cmdclass(),
    install_requires=read_requirements_file("base"),
    extras_require={
        "rllib": read_requirements_file("rllib"),
    },
    entry_points={
        "console_scripts": [
            "neural-mmo-forge=neural_mmo.Forge:main",
        ]
    },
    python_requires=">=3.6",
    license="MIT",
    author="Joseph Suarez",
    author_email="sealsuarez@gmail.com",
    url=REPO_URL,
    keywords=["Neural MMO", "MMO"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    include_package_data=True,
)




from pdb import set_trace as T

from neural_mmo.forge.blade import core
from neural_mmo.forge.blade.core import config


class RLlibConfig:
   '''Base config for RLlib Models

   Extends core Config, which contains environment, evaluation,
   and non-RLlib-specific learning parameters'''

   @property
   def MODEL(self):
      return self.__class__.__name__
  
   #Hardware Scale
   NUM_GPUS_PER_WORKER     = 0
   NUM_GPUS                = 1
   NUM_WORKERS             = 1
   LOCAL_MODE              = False
   LOAD                    = True

   #Memory/Batch Scale
   TRAIN_EPOCHS            = 10000
   LSTM_BPTT_HORIZON       = 16
   NUM_SGD_ITER            = 1

   #Model
   SCRIPTED                = None
   N_AGENT_OBS             = 100
   NPOLICIES               = 1
   HIDDEN                  = 64
   EMBED                   = 64

   #Reward
   TEAM_SPIRIT             = 0.0
   ACHIEVEMENT_SCALE       = 1.0/15.0


class LargeMaps(core.Config, RLlibConfig, config.AllGameSystems):
   '''Large scale Neural MMO training setting

   Features up to 1000 concurrent agents and 1000 concurrent NPCs,
   1km x 1km maps, and 5/10k timestep train/eval horizons

   This is the default setting as of v1.5 and allows for large
   scale multiagent research even on relatively modest hardware'''

   #Memory/Batch Scale
   NUM_WORKERS             = 1 #16
   TRAIN_BATCH_SIZE        = 32 * NUM_WORKERS #Bug? This gets doubled
   ROLLOUT_FRAGMENT_LENGTH = 32
   SGD_MINIBATCH_SIZE      = 256

   #Horizon
   TRAIN_HORIZON           = 8192
   EVALUATION_HORIZON      = 8192


class SmallMaps(config.SmallMaps, RLlibConfig, config.AllGameSystems):
   '''Small scale Neural MMO training setting

   Features up to 128 concurrent agents and 32 concurrent NPCs,
   60x60 maps (excluding the border), and 1000 timestep train/eval horizons.
   
   This setting is modeled off of v1.1-v1.4 It is appropriate as a quick train
   task for new ideas, a transfer target for agents trained on large maps,
   or as a primary research target for PCG methods.'''

   #Memory/Batch Scale
   NUM_WORKERS             = 6 #32
   TRAIN_BATCH_SIZE        = 256 * NUM_WORKERS #Bug? This gets doubled
   ROLLOUT_FRAGMENT_LENGTH = 256
   SGD_MINIBATCH_SIZE      = min(256, TRAIN_BATCH_SIZE)
 
   #Horizon
   TRAIN_HORIZON           = 1024
   EVALUATION_HORIZON      = 1024


class Debug(SmallMaps, config.AllGameSystems):
   '''Debug Neural MMO training setting

   A version of the SmallMap setting with greatly reduced batch parameters.
   Only intended as a tool for identifying bugs in the model or environment'''
   LOAD                    = False
   LOCAL_MODE              = True
   NUM_WORKERS             = 1

   SGD_MINIBATCH_SIZE      = 100
   TRAIN_BATCH_SIZE        = 400
   TRAIN_HORIZON           = 200
   EVALUATION_HORIZON      = 50

   HIDDEN                  = 2
   EMBED                   = 2


### AICrowd competition settings
class Competition(config.AllGameSystems, config.Achievement): pass
class CompetitionRound1(SmallMaps, Competition):
   @property
   def SPAWN(self):
      return self.SPAWN_CONCURRENT

   NENT                    = 128
   NPOP                    = 1

class CompetitionRound2(SmallMaps, Competition):
   @property
   def SPAWN(self):
      return self.SPAWN_CONCURRENT

   NENT                    = 128
   NPOP                    = 16
   COOPERATIVE             = True

class CompetitionRound3(LargeMaps, Competition):
   @property
   def SPAWN(self):
      return self.SPAWN_CONCURRENT

   NENT                    = 1024
   NPOP                    = 32
   COOPERATIVE             = True


### NeurIPS Experiments
class SmallMultimodalSkills(SmallMaps, config.AllGameSystems): pass
class LargeMultimodalSkills(LargeMaps, config.AllGameSystems): pass


class MagnifyExploration(SmallMaps, config.Resource, config.Progression):
   pass
class Population4(MagnifyExploration):
   NENT  = 256
class Population32(MagnifyExploration):
   NENT  = 256
class Population256(MagnifyExploration):
   NENT  = 256


class DomainRandomization16384(SmallMaps, config.AllGameSystems):
   TERRAIN_TRAIN_MAPS=16384
class DomainRandomization256(SmallMaps, config.AllGameSystems):
   TERRAIN_TRAIN_MAPS=256
class DomainRandomization32(SmallMaps, config.AllGameSystems):
   TERRAIN_TRAIN_MAPS=32
class DomainRandomization1(SmallMaps, config.AllGameSystems):
   TERRAIN_TRAIN_MAPS=1


class TeamBased(MagnifyExploration, config.Combat):
   NENT                    = 128
   NPOP                    = 32
   COOPERATIVE             = True
   TEAM_SPIRIT             = 0.5

   @property
   def SPAWN(self):
      return self.SPAWN_CONCURRENT



from collections import defaultdict

class MultiSet:
   def __init__(self, capacity=0):
      self.data = defaultdict(int)
      self.count = 0
      self.capacity = capacity

   def __iter__(self):
      return self.data.__iter__()

   def countItem(self, item):
      return self.data[item]

   @property
   def full(self):
      return self.capacity != 0 and self.count >= self.capacity

   @property
   def empty(self):
      return self.count == 0

   def get(self, item):
      return self.data[item]

   def isIn(self, item, num=1):
      return self.data[item] > num

   def add(self, item, num=1):
      assert self.capacity == 0 or self.count+num <= self.capacity
      self.data[item] += num
      self.count += num
   
   #Alias
   def remove(self, item, num=1):
      self.pop(item, num)

   def pop(self, item, num=1):
      assert self.capacity == 0 or self.count-num >= 0
      self.data[item] -= num
      self.count -= num

   def union(self, other):
      for e in other:
         self.add(e, other.count(e))

   def diff(self, other):
      for e in other:
         self.remove(e, other.count(e))

   def contains(self, other):
      for e in other:
         if not isIn(e, other[e]):
            return False
      return True




from pdb import set_trace as T
import numpy as np

import time
import os
 
from collections import defaultdict

import ray

from neural_mmo.forge.blade.systems import visualizer
from neural_mmo.forge.blade.lib.enums import Neon, Solid

from tornado.ioloop import IOLoop
import bokeh, jinja2

#Adapted from https://rigtorp.se/2011/01/01/rolling-statistics-numpy.html
def rolling_window(a, window, pad=True):
   a = np.array(a)
   if pad:
      a = np.pad(a, pad_width=(window-1, 0), mode='edge')
   shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
   strides = a.strides + (a.strides[-1],)
   return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

def polar(r, frac): 
   angle = 2 * 3.1415927 * frac
   return r*np.sin(angle), r*np.cos(angle)
 
def flat(vals):
   flat = []
   for key, val in sorted(vals.items()):
      flat += val
   return flat

def disableAxis(axis):
   axis.major_label_text_font_size = '0pt'
   axis.major_tick_line_color      = None
   axis.minor_tick_line_color      = None
   axis.axis_line_color            = None

def blank(width, height, blob={}, ylabel=None, colors=[]):
   fig = bokeh.plotting.figure(
         plot_width=width,
         plot_height=height,
         toolbar_location=None)

   fig.grid[0].visible = False
   fig.grid[1].visible = False

   line = fig.line([0, 0], [0, 0])
   line.visible = False

   disableAxis(fig.xaxis)
   disableAxis(fig.yaxis)

   legend = {}
   for idx, (color, (key, source)) in enumerate(zip(colors, blob.items())):
      if key is None:
         key = ylabel
      glyph = fig.circle([0], [0], name=key, size=0, color=color)
      legend[key] = [glyph]

   return fig, legend

class Plot:
   def __init__(self, config, fig, key, n):
      self.config = config
      self.fig    = fig
      self.key    = key
      self.n      = n

   def render(self, blob, colors):
      legend = []
      for idx, (color, (key, source)) in enumerate(zip(colors, blob.items())):
         if key is None:
            key = self.key

         legend.append(self.glyphs(source, key, color, idx))

      return legend

   def glyphs(self, source, key, color, idx):
      source = self.preprocess(source, key, color, idx)
      source = bokeh.plotting.ColumnDataSource(source)
      return self.plot(source, key, color, idx)

class MarginPlot(Plot):
   def __init__(self, config, fig, key, n, margin=0.05):
      super().__init__(config, fig, key, n)
      self.fig.y_range.range_padding = margin
      self.fig.x_range.range_padding = margin
 
class Training(Plot):
   def preprocess(self, data, key, color, idx, w=20):
      mean                   = np.array(data['Mean'])
      std                    = np.array(data['Std'])

      preprocessed           = {}
      preprocessed['x']      = data['x']
      preprocessed['smooth'] = mean
      preprocessed['lower']  = mean - std
      preprocessed['upper']  = mean + std

      return preprocessed

   def plot(self, source, key, color, idx):
      self.fig.xaxis.axis_label = 'Environment Episodes'
      self.fig.yaxis.axis_label = 'Normalized Value'

      line = self.fig.line(
         source=source,
         name=key,
         x='x',
         y='smooth',
         color=color)

      return [line]


class Line(Plot):
   def preprocess(self, data, key, color, idx, w=20):
      sliding = rolling_window(flat(data), w)
      smooth  = np.mean(sliding, axis=1)
      std     = np.std(sliding, axis=1)

      preprocessed           = {}
      preprocessed['x']      = np.arange(len(smooth))
      preprocessed['smooth'] = smooth
      preprocessed['lower']  = smooth - std
      preprocessed['upper']  = smooth + std

      return preprocessed

   def plot(self, source, key, color, idx):
      self.fig.xaxis.axis_label = 'Index in Logs'
      self.fig.yaxis.axis_label = 'Value'

      band = self.fig.varea(
         source=source,
         name=key,
         x='x',
         y1='lower',
         y2='upper',
         fill_color=color)

      lower = self.fig.line(
         source=source,
         name=key,
         x='x',
         y='lower',
         line_color=color,
         line_width=0.5,
         line_alpha=0.7,
         color=color)

      upper = self.fig.line(
         source=source,
         name=key,
         x='x',
         y='upper',
         line_color=color,
         line_width=0.5,
         line_alpha=0.7,
         color=color)

      line = self.fig.line(
         source=source,
         name=key,
         x='x',
         y='smooth',
         color=color)

      return [band, lower, upper, line]

class Scatter(MarginPlot):
   def preprocess(self, data, key, color, idx, w=20):
      preprocessed = defaultdict(list)

      for t, vList in data.items():
         for val in vList:
            preprocessed['x'].append(t)
            preprocessed['y'].append(val)

      return preprocessed

   def plot(self, source, key, color, idx):
      self.fig.xaxis.axis_label = 'Game Tick'
      self.fig.yaxis.axis_label = 'Value'

      circle = self.fig.circle(
         source=source,
         name=key,
         x='x',
         y='y',
         color=color)

      return [circle]


class Histogram(Plot):
   def preprocess(self, data, key, color, idx, nBins=25):
      hist, edges = np.histogram(flat(data), bins=nBins)
         
      preprocessed          = {}
      preprocessed['top']   = hist
      preprocessed['left']  = edges[:-1]
      preprocessed['right'] = edges[1:]
 
      return preprocessed

   def plot(self, source, key, color, idx):
      self.fig.xaxis.axis_label = 'Value'
      self.fig.yaxis.axis_label = 'Count'

      quad = self.fig.quad(
         source=source,
         name=key,
         top='top',
         bottom=0,
         left='left',
         right='right',
         fill_color=color,
         line_color=color,
         )

      return [quad]

class Gantt(Plot):
   def preprocess(self, data, key, color, idx):
      preprocessed = []
      
      for tick, vList in sorted(data.items()):
         for val in vList:
            preprocessed.append((tick-val, tick))

      preprocessed = sorted(preprocessed)
      bottom, top  = list(zip(*preprocessed))
      bottom, top  = np.array(bottom), np.array(top)

      preprocessed           = {}
      preprocessed['top']    = top
      preprocessed['bottom'] = bottom
      preprocessed['left']   = bottom - 1
      preprocessed['right']  = bottom

      return preprocessed

   def plot(self, source, key, color, idx):
      self.fig.xaxis.axis_label = 'Game Tick'
      self.fig.yaxis.axis_label = 'Range'

      quad = self.fig.quad(
         source=source,
         name=key,
         top='top',
         bottom='bottom',
         left='left',
         right='right',
         fill_color=color,
         line_color=None,
         )

      return [quad]

class Stats(Plot):
   def preprocess(self, data, key, color, idx):
      data  = flat(data)
      mmean = np.mean(data)
      mmin  = np.min(data)
      mmax  = np.max(data)

      preprocessed           = {}
      preprocessed['top']    = [mmax]
      preprocessed['mean']   = [mmean]
      preprocessed['bottom'] = [mmin]
      preprocessed['left']   = [idx]
      preprocessed['right']  = [idx+1]

      return preprocessed

   def plot(self, source, key, color, idx):
      self.fig.xaxis.axis_label = 'Value'
      self.fig.yaxis.axis_label = 'Stat'
      disableAxis(self.fig.xaxis)

      quad = self.fig.quad(
         source=source,
         name=key,
         top='top',
         bottom='bottom',
         left='left',
         right='right',
         fill_color=color,
         line_color=color,
         )

      quad = self.fig.quad(
         source=source,
         name=key,
         top='mean',
         bottom='mean',
         left='left',
         right='right',
         fill_color=color,
         line_color=color,
         line_width=3.0,
         fill_alpha=1.0
         )


      return [quad]

class Radar(MarginPlot):
   def preprocess(self, data, key, color, idx):
      data  = flat(data)
      mmean = np.mean(data)
      mmin  = np.min(data)
      mmax  = np.max(data)

      preprocessed           = {}
      preprocessed['top']    = [mmax]
      preprocessed['mean']   = [mmean]
      preprocessed['bottom'] = [mmin]
      preprocessed['left']   = [idx]
      preprocessed['right']  = [idx+1]

      return preprocessed


   def render(self, blob, colors):
       self.fig.xaxis.axis_label = 'Radar Axis'
       self.fig.yaxis.axis_label = 'Radar Axis'
       disableAxis(self.fig.xaxis)
       disableAxis(self.fig.yaxis)

       legend, data = {}, defaultdict(list)
       for idx, (color, (subkey, source)) in enumerate(zip(colors, blob.items())):
          for key, val in self.preprocess(source, None, color, idx).items():
             data[key] += val

       n = len(blob)

       radians = 2 * 3.14159265

       #data['top']    = np.array(data['top'])*0 + 10
       #data['mean']   = np.array(data['mean'])*0 + 5
       #data['bottom'] = np.array(data['bottom'])*0 + 2

       for idx, (color, (subkey, source)) in enumerate(zip(colors, blob.items())):
          if subkey is None:
             subkey = key


          lIdx   = (idx - 1) % n
          mIdx   = idx % n
          rIdx   = (idx + 1) % n

          xlb, ylb = polar(data['bottom'][lIdx],  lIdx/n)
          xmb, ymb = polar(data['bottom'][mIdx],  mIdx/n)
          xrb, yrb = polar(data['bottom'][rIdx], rIdx/n)

          xlm, ylm = polar(data['mean'][lIdx],  lIdx/n)
          xmm, ymm = polar(data['mean'][mIdx],  mIdx/n)
          xrm, yrm = polar(data['mean'][rIdx],  rIdx/n)

          xlt, ylt = polar(data['top'][lIdx],  lIdx/n)
          xmt, ymt = polar(data['top'][mIdx],  mIdx/n)
          xrt, yrt = polar(data['top'][rIdx], rIdx/n)


          xlb, ylb = (xlb+xmb)/2, (ylb+ymb)/2
          xrb, yrb = (xmb+xrb)/2, (ymb+yrb)/2

          xlm, ylm = (xlm+xmm)/2, (ylm+ymm)/2
          xrm, yrm = (xmm+xrm)/2, (ymm+yrm)/2

          xlt, ylt = (xlt+xmt)/2, (ylt+ymt)/2
          xrt, yrt = (xmt+xrt)/2, (ymt+yrt)/2


          x = [xrb, xmb, xlb, xlt, xmt, xrt]
          y = [yrb, ymb, ylb, ylt, ymt, yrt]
          source = {'x': x, 'y': y}
          source = bokeh.plotting.ColumnDataSource(source)
          
          quad = self.fig.patch(
            source=source,
            name=key,
            x='x',
            y='y',
            fill_color=color,
            line_color=color,
            line_width=0.0,
            )

          x = [xlm, xmm, xrm]
          y = [ylm, ymm, yrm]
          source = {'x': x, 'y': y}
          source = bokeh.plotting.ColumnDataSource(source)
 
          mean = self.fig.line(
            source=source,
            name=key,
            x='x',
            y='y',
            line_width=2.5,
            color=color,)

          x = [xlb, xmb, xrb]
          y = [ylb, ymb, yrb]
          source = {'x': x, 'y': y}
          source = bokeh.plotting.ColumnDataSource(source)
 
          bottom = self.fig.line(
            source=source,
            name=key,
            x='x',
            y='y',
            line_width=2.5,
            color=color,)

          x = [xlt, xmt, xrt]
          y = [ylt, ymt, yrt]
          source = {'x': x, 'y': y}
          source = bokeh.plotting.ColumnDataSource(source)
 
          top = self.fig.line(
            source=source,
            name=key,
            x='x',
            y='y',
            line_width=2.5,
            color=color,)


          legend[subkey] = [quad, mean, bottom, top]

       return legend

class StackedArea(Plot):
   def preprocess(self, data, key, color, idx, w=20):
      sliding = rolling_window(flat(data), w)
      smooth  = np.mean(sliding, axis=1)
      std     = np.std(sliding, axis=1)

      preprocessed           = {}
      preprocessed['x']      = np.arange(len(smooth))
      preprocessed['smooth'] = smooth
      preprocessed['lower']  = smooth - std
      preprocessed['upper']  = smooth + std

      return preprocessed

   def render(self, blob, colors):
       self.fig.xaxis.axis_label = 'Index in Logs'
       self.fig.yaxis.axis_label = 'Value'

       legend, bottom = {}, None
       for idx, (color, (subkey, source)) in enumerate(zip(colors, blob.items())):
          if subkey is None:
             subkey = self.key

          data = self.preprocess(source, None, color, idx)

          x = np.array(data['x'])
          y = np.array(data['smooth'])

          if bottom is None:
             bottom = x * 0.0
        
          data   = {'x': x, 'y1': bottom, 'y2': bottom + y}
          source = bokeh.plotting.ColumnDataSource(data)
          bottom = bottom + y
          
          band = self.fig.varea(
             source=source,
             name=subkey,
             x='x',
             y1='y1',
             y2='y2',
             fill_color=color)

          lower = self.fig.line(
             source=source,
             name=subkey,
             x='x',
             y='y1',
             line_color=color,
             line_width=0.5,
             color=color)

          upper = self.fig.line(
             source=source,
             name=subkey,
             x='x',
             y='y2',
             line_color=color,
             line_width=0.5,
             color=color)

          legend[subkey] = [band, lower, upper]

       return legend



import numpy as np
import torch
from torch import optim
from torch.autograd import Variable

from collections import defaultdict
from neural_mmo.forge.ethyr.torch import loss


class ManualAdam(optim.Adam):
   '''Adam wrapper that accepts gradient lists'''
   def step(self, grads):
      '''Takes an Adam step on the parameters using
      the user provided gradient list provided.
   
      Args:
         grads: A list of gradients
      '''
      grads = Variable(torch.Tensor(np.array(grads)))
      self.param_groups[0]['params'][0].grad = grads
      super().step()

class ManualSGD(optim.SGD):
   '''SGD wrapper that accepts gradient lists'''
   def step(self, grads):
      '''Takes an SGD step on the parameters using
      the user provided gradient list provided.
   
      Args:
         grads: A list of gradients
      '''
      grads = Variable(torch.Tensor(np.array(grads)))
      self.param_groups[0]['params'][0].grad = grads
      super().step()

def merge(rollouts):
   '''Merges all collected rollouts for batched
   compatibility with optim.backward'''
   n = 0
   outs = defaultdict(lambda: defaultdict(list))
   for rollout in rollouts.values():
      for idx in range(rollout.time):
         for out in rollout.actions[idx]:
            if len(out.atnLogits) == 1:
               continue
            outk = outs[out.atnArgKey]
            outk['atns'].append(out.atnLogits)
            outk['idxs'].append(out.atnIdx)
            outk['vals'].append(out.value)
            outk['rets'].append(out.returns)
            n += 1

   return outs, n


def backward(rollouts, config):
   '''Computes gradients from a list of rollouts

   Args:
      rolls: A list of rollouts
      valWeight (float): Scale to apply to the value loss
      entWeight (float): Scale to apply to the entropy bonus
      device (str): Hardware to run backward on

   Returns:
      reward: Mean reward achieved across rollouts
      val: Mean value function estimate across rollouts
      pg: Policy gradient loss
      valLoss: Value loss
      entropy: Entropy bonus      
   '''
   device = config.DEVICE
   outs, n = merge(rollouts)
   pgLoss, valLoss, entLoss = 0, 0, 0
   for k, out in outs.items():
      atns = out['atns']
      vals = torch.stack(out['vals'])
      idxs = torch.tensor(out['idxs']).to(device)
      rets = torch.tensor(out['rets']).view(-1, 1).to(device)

      l, v, e = loss.PG(atns, idxs, vals, rets)

      #Averaging results in no learning. Need to retune LR?
      pgLoss  += l# / n
      valLoss += v# / n
      entLoss += e# / n

   totLoss = (
         config.PG_WEIGHT*pgLoss + 
         config.VAL_WEIGHT*valLoss + 
         config.ENTROPY*entLoss)

   totLoss.backward(retain_graph=True)

   return pgLoss, valLoss, entLoss

